from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import json
import logging
import numpy as np
import os
import pathlib
import pytransform3d.rotations as ptr
import pyvista as pv
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.Dock import Dock, DockArea
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView
from NaviNIBS.Navigator.GUI.Widgets.SurfViews import Surf3DView
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import FullTargetsTableWidget
from NaviNIBS.Navigator.GUI.Widgets.EditTargetWidget import EditTargetWidget
from NaviNIBS.Navigator.GUI.Widgets.EditGridWidget import EditGridWidget
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QueuedRedrawMixin import QueuedRedrawMixin
from NaviNIBS.util.pyvista import Actor, RemotePlotterProxy
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import composeTransform, applyTransform, invertTransform, concatenateTransforms
from NaviNIBS.Navigator.Model.Session import Session, Target


logger = logging.getLogger(__name__)


@attrs.define
class VisualizedTarget:
    """
    Note: this doesn't connect to any change signals from `target`, instead assuming that caller will
    re-instantiate the `VisualizedTarget` for any target changes.
    """
    _target: Target
    _plotter: DefaultBackgroundPlotter = attrs.field(repr=False)
    _style: str
    _color: str = '#2222FF'
    _actors: tp.Dict[str, Actor] = attrs.field(init=False, factory=dict, repr=False)
    _visible: bool = True  # track this separately from self._target.isVisible to allow temporarily overriding

    def __attrs_post_init__(self):
        self.plot()

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, isVisible: bool):
        if self._visible == isVisible:
            return

        if isVisible:
            for actor in self._actors.values():
                actor.VisibilityOn()
        else:
            for actor in self._actors.values():
                actor.VisibilityOff()

        self._visible = isVisible

        with self._plotter.allowNonblockingCalls():
            self._plotter.render()

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style: str):
        if self._style == style:
            return

        self.clearActors()
        self._style = style
        self.plot()

    def plot(self):
        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not ready yet
            return

        thinWidth = 3
        thickWidth = 6
        if self._style == 'line':
            pts_line = np.vstack((self._target.entryCoord, self._target.targetCoord))
            self._actors['line'] = self._plotter.add_lines(pts_line,
                                                           color=self._color,
                                                           width=thickWidth,
                                                           label=self._target.key,
                                                           name=self._target.key + 'line',
                                                           )

        elif self._style == 'lines':
            pts_line1 = np.vstack((self._target.entryCoord, self._target.targetCoord))
            self._actors['line1'] = self._plotter.add_lines(pts_line1,
                                                           color=self._color,
                                                           width=thickWidth,
                                                           label=self._target.key,
                                                           name=self._target.key + 'line1',
                                                           )

            pts_line2 = applyTransform(self._target.coilToMRITransf, np.asarray([[0, -10, 0], [0, 0, 0]]))
            self._actors['line2'] = self._plotter.add_lines(pts_line2,
                                                            color=self._color,
                                                            width=thinWidth,
                                                            label=self._target.key,
                                                            name=self._target.key + 'line2',
                                                            )
            pts_line3 = applyTransform(self._target.coilToMRITransf, np.asarray([[0, 0, -50], [0, 0, 10]]))
            self._actors['line3'] = self._plotter.add_lines(pts_line3,
                                                            color=self._color,
                                                            width=thinWidth,
                                                            label=self._target.key,
                                                            name=self._target.key + 'line3',
                                                            )

            self._actors['target'] = self._plotter.add_points(self._target.targetCoord,
                                                              color=self._color,
                                                              point_size=10.,
                                                              render_points_as_spheres=True,
                                                              label=self._target.key,
                                                              name=self._target.key + 'target')

        elif self._style == 'coilLines':
            coilDiameter = 10
            coilHandleLength = 7

            theta = np.linspace(0, 2*np.pi, 361)
            circlePts = np.column_stack((coilDiameter/2 * np.cos(theta), coilDiameter/2 * np.sin(theta), np.zeros(theta.shape)))

            for wing, dir in (('wing1', 1.), ('wing2', -1.)):
                if self._target.coilToMRITransf is not None:
                    pts_wing = applyTransform(self._target.coilToMRITransf, circlePts + dir*np.asarray([coilDiameter/2, 0, 0]))
                    self._actors[wing] = self._plotter.add_lines(pts_wing,
                                                                 connected=True,
                                                                 color=self._color,
                                                                 width=thinWidth,
                                                                 label=self._target.key,
                                                                 name=self._target.key + wing,
                                                                 )

            if self._target.coilToMRITransf is not None:
                pts_line2 = applyTransform(self._target.coilToMRITransf, np.asarray([[0, -coilHandleLength, 0], [0, 0, 0]]))
                self._actors['line2'] = self._plotter.add_lines(pts_line2,
                                                                color=self._color,
                                                                width=thinWidth,
                                                                label=self._target.key,
                                                                name=self._target.key + 'line2',
                                                                )
                if self._target.targetCoord is not None and self._target.entryCoord is not None:
                    depth = np.linalg.norm(self._target.targetCoord - self._target.entryCoord) + self._target.depthOffset
                else:
                    depth = 50
                pts_line3 = applyTransform(self._target.coilToMRITransf, np.asarray([[0, 0, -depth], [0, 0, 10]]))
                self._actors['line3'] = self._plotter.add_lines(pts_line3,
                                                                color=self._color,
                                                                width=thinWidth,
                                                                label=self._target.key,
                                                                name=self._target.key + 'line3',
                                                                )
            elif self._target.targetCoord is not None and self._target.entryCoord is not None:
                pts_line2 = np.vstack([self._target.targetCoord, self._target.entryCoord])
                self._actors['line2'] = self._plotter.add_lines(pts_line2,
                                                                color=self._color,
                                                                width=thinWidth,
                                                                label=self._target.key,
                                                                name=self._target.key + 'line2',
                                                                )

            self._actors['target'] = self._plotter.add_points(self._target.targetCoord,
                                                              color=self._color,
                                                              point_size=10.,
                                                              render_points_as_spheres=True,
                                                              label=self._target.key,
                                                              name=self._target.key + 'target',
                                                              reset_camera=False,
                                                              render=False)

        else:
            raise NotImplementedError()

        with self._plotter.allowNonblockingCalls():
            if not self.visible:
                for actor in self._actors.values():
                    actor.VisibilityOff()

            self._plotter.render()

    @property
    def actors(self):
        return self._actors

    def clearActors(self):
        with self._plotter.allowNonblockingCalls():
            for actor in self._actors.values():
                self._plotter.remove_actor(actor)
            self._actors.clear()
            self._plotter.render()


@attrs.define
class TargetsPanel(MainViewPanelWithDockWidgets, QueuedRedrawMixin):
    _key: str = 'Set targets'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-flash-outline'))
    _tableWdgt: FullTargetsTableWidget = attrs.field(init=False)
    _views: tp.Dict[str, tp.Union[MRISliceView, Surf3DView]] = attrs.field(init=False, factory=dict)
    _targetActors: tp.Dict[str, VisualizedTarget] = attrs.field(init=False, factory=dict)

    _editTargetWdgt: EditTargetWidget = attrs.field(init=False)
    _editGridWdgt: EditGridWidget = attrs.field(init=False)
    _editGridDock: Dock = attrs.field(init=False)

    _targetDispStyle_comboBox: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)

    _surfKeys: tp.List[str] = attrs.field(factory=lambda: ['gmSurf', 'skinSurf'])

    _enabledOnlyWhenTargetSelected: list[QtWidgets.QWidget | EditTargetWidget] = attrs.field(init=False, factory=list)

    _redrawTargetKeys: set[str] = attrs.field(init=False, factory=set)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        MainViewPanelWithDockWidgets.__attrs_post_init__(self)
        QueuedRedrawMixin.__attrs_post_init__(self)

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        if not self.session.headModel.isSet:
            return False, 'No head model set'
        return True, None

    @staticmethod
    def _getRotMatForCoilAxis(axis: str) -> np.ndarray:
        if axis == 'x':
            return ptr.active_matrix_from_extrinsic_euler_yxy([-np.pi/2, np.pi/2, 0])
        elif axis == 'y':
            return ptr.active_matrix_from_angle(0, np.pi/2)
        elif axis in ('z', '3D'):
            return np.eye(3)
        else:
            raise NotImplementedError()

    def _finishInitialization(self):
        # don't initialize computationally-demanding views until panel is activated (viewed)

        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QHBoxLayout())

        dock, container = self._createDockWidget(
            title='Targets',
            layout=QtWidgets.QVBoxLayout()
        )
        dock.setStretch(1, 10)
        container.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self._wdgt.addDock(dock, position='left')

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        container.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Import targets from file...')
        btn.clicked.connect(self._onImportTargetsBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 0, 1, 2)

        btn = QtWidgets.QPushButton('Add target')
        btn.clicked.connect(self._onAddBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 0)

        btn = QtWidgets.QPushButton('Delete target')
        btn.clicked.connect(self._onDeleteBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 1)
        self._enabledOnlyWhenTargetSelected.append(btn)

        btn = QtWidgets.QPushButton('Duplicate target')
        btn.clicked.connect(self._onDuplicateBtnClicked)
        btnContainer.layout().addWidget(btn, 2, 0)
        self._enabledOnlyWhenTargetSelected.append(btn)

        btn = QtWidgets.QPushButton('Goto target')
        btn.clicked.connect(self._onGotoBtnClicked)
        btnContainer.layout().addWidget(btn, 2, 1)
        self._enabledOnlyWhenTargetSelected.append(btn)

        self._tableWdgt = FullTargetsTableWidget()
        self._tableWdgt.sigCurrentItemChanged.connect(self._onCurrentTargetChanged)
        self._tableWdgt.sigSelectionChanged.connect(self._onSelectionChanged)
        container.layout().addWidget(self._tableWdgt.wdgt)

        fieldContainer = QtWidgets.QWidget()
        fieldLayout = QtWidgets.QFormLayout()
        fieldContainer.setLayout(fieldLayout)
        container.layout().addWidget(fieldContainer)
        self._targetDispStyle_comboBox.addItems(['line', 'lines', 'coilLines'])
        self._targetDispStyle_comboBox.setCurrentIndex(2)
        self._targetDispStyle_comboBox.currentIndexChanged.connect(self._onTargetDispStyleChanged)
        fieldLayout.addRow('Target display style:', self._targetDispStyle_comboBox)

        self._editTargetWdgt = EditTargetWidget(session=self.session,
                                                wdgt=QtWidgets.QWidget(),
                                                getNewTargetCoord=self._getCrosshairCoord,
                                                setTargetCoordButtonLabel='Set from crosshair position',
                                                getNewEntryCoord=self._getCrosshairCoord,
                                                setEntryCoordButtonLabel='Set from crosshair position',
                                                )
        dock, _ = self._createDockWidget(
            title='Edit target',
            widget=self._editTargetWdgt.wdgt,
        )
        dock.setStretch(1, 10)
        self._wdgt.addDock(dock, position='bottom')
        dock_editTarget = dock

        self._editGridWdgt = EditGridWidget(session=self.session,
                                            wdgt=QtWidgets.QWidget(),
                                            )
        dock, _ = self._createDockWidget(
            title='Edit grid',
            widget=self._editGridWdgt.wdgt,
        )
        dock.setStretch(1, 10)
        self._editGridDock = dock
        if False:
            self._wdgt.addDock(dock, position='bottom')
        else:
            self._wdgt.addDock(dock, position='below', relativeTo=dock_editTarget)

        dock, container = self._createDockWidget(
            title='Target views',
            layout=QtWidgets.QGridLayout()
        )
        self._wdgt.addDock(dock, position='right')

        # TODO: put each view in its own dockwidget instead of a grid layout so they can be moved around

        for iRow, iCol, key in ((0, 1, 'x'), (0, 0, 'y'), (1, 0, 'z'), (1, 1, '3D')):
            if key in ('x', 'y', 'z'):
                self._views[key] = MRISliceView(label=key, normal=self._getRotMatForCoilAxis(key))
            elif key == '3D':
                self._views[key] = Surf3DView(label=key, normal=np.eye(3),
                                              activeSurf=self._surfKeys,
                                              pickableSurfs=[self._surfKeys[0]],  # assuming surfKeys are [gm, skin], make only gm pickable
                                              cameraOffsetDist=50,
                                              surfOpacity=[0.8, 0.5])
            else:
                raise NotImplementedError()

            self._views[key].sigSliceTransformChanged.connect(lambda key=key: self._onSliceTransformChanged(sourceKey=key))

            container.layout().addWidget(self._views[key].wdgt, iRow, iCol)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        for widget in self._enabledOnlyWhenTargetSelected:
            logger.debug(f'Disabling {widget}')
            widget.setEnabled(False)

        self._onTargetsChanged()

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for view in self._views.values():
            if isinstance(view.plotter, RemotePlotterProxy):
                await view.plotter.isReadyEvent.wait()

        self._onTargetsChanged()

        self.finishedAsyncInit.set()

    def _getCrosshairCoord(self) -> np.ndarray:
        """
        Used by CoordinateWidgets to set target or entry coord from current crosshair position
        """
        return self._views['3D'].sliceOrigin

    def _onTargetDispStyleChanged(self, index: int):
        for visualizedTarget in self._targetActors.values():
            visualizedTarget.style = self._targetDispStyle_comboBox.currentText()

    def _onSliceTransformChanged(self, sourceKey: str):
        logger.debug('Slice {} transform changed'.format(sourceKey))
        crossTransf = self._views[sourceKey].sliceTransform @ composeTransform(np.linalg.pinv(self._getRotMatForCoilAxis(axis=sourceKey)))  # TODO: double check order
        for key, view in self._views.items():
            if key == sourceKey:
                continue
            logger.debug('Updating {} slice transform'.format(key))
            view.sliceTransform = crossTransf @ composeTransform(self._getRotMatForCoilAxis(axis=key))

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self.session.targets.sigItemsChanged.connect(self._onTargetsChanged)
        self._tableWdgt.session = self.session

        for key, view in self._views.items():
            view.session = self.session

    def _getCurrentTargetKey(self) -> tp.Optional[str]:
        return self._tableWdgt.currentCollectionItemKey

    def _onSelectionChanged(self, keys: list[str]):
        for widget in self._enabledOnlyWhenTargetSelected:
            logger.debug(f"{'Disabling' if len(keys)==0 else 'Enabling'} {widget}")
            widget.setEnabled(len(keys)>0)

    def _onCurrentTargetChanged(self, targetKey: str | None):
        if targetKey is not None:
            self._gotoTarget(targetKey=targetKey)    # go to target immediately whenever selection changes

    def _gotoTarget(self, targetKey: str):

        logger.debug('gotoTarget')

        plotter = self._views['3D'].plotter
        if isinstance(plotter, RemotePlotterProxy) and not plotter.isReadyEvent.is_set():
            return  # plotter not ready yet

        # change slice camera views to align with selected target
        target = self.session.targets[targetKey]
        extraTransf = np.eye(4)
        if True:
            # align at target depth
            extraTransf[:3, 3] = applyTransform(invertTransform(target.coilToMRITransf), target.targetCoord, doCheck=False)
        else:
            # align at coil depth
            pass

        self._views['3D'].sliceTransform = concatenateTransforms([
            composeTransform(self._getRotMatForCoilAxis(axis='3D')),
            extraTransf,
            target.coilToMRITransf])

        if True:
            # also set camera position for 3D view to align with target
            plotter = self._views['3D'].plotter
            with plotter.allowNonblockingCalls():
                plotter.camera.focal_point = self._views['3D'].sliceOrigin
                cameraPos = applyTransform(self._views['3D'].sliceTransform, np.asarray([0, 0, 200]))
                plotter.camera.position = cameraPos
                plotter.camera.up = applyTransform(self._views['3D'].sliceTransform,
                                                                 np.asarray([0, 100, 0])) - cameraPos
                plotter.render()

    def _createVisualForTarget(self, viewKey: str, target: Target):
        logger.debug(f'Creating VisualizedTarget for target {target.asDict()} {target.coilToMRITransf}')
        view = self._views[viewKey]
        style = self._targetDispStyle_comboBox.currentText()
        self._targetActors[viewKey + target.key] = VisualizedTarget(target=target,
                                                                    plotter=view.plotter,
                                                                    style=style,
                                                                    visible=target.isVisible)

    def _onTargetsChanged(self, changedTargetKeys: tp.Optional[tp.List[str]] = None, changedTargetAttrs: tp.Optional[tp.List[str]] = None):
        if not self._hasInitialized and not self._isInitializing:
            return

        # TODO: move this below to only delay visualization initialization, not target table, etc.
        for view in self._views.values():
            if isinstance(view.plotter, RemotePlotterProxy) and not view.plotter.isReadyEvent.is_set():
                # plotter not ready yet
                return

        logger.debug(f'Targets changed. Updating tree view and plots.\nchangedTargetKeys={changedTargetKeys}, \nchangedTargetAttrs={changedTargetAttrs}')

        if changedTargetKeys is None:
            changedTargetKeys = self.session.targets.keys()

        if changedTargetAttrs == ['isVisible']:
            # only visibility changed

            for targetKey in changedTargetKeys:
                target = self._session.targets[targetKey]
                for viewKey in self._views:
                    actorKey = viewKey + targetKey
                    if actorKey in self._targetActors:
                        self._targetActors[actorKey].visible = target.isVisible
                    elif target.isVisible:
                        self._redrawTargetKeys.add(targetKey)
                        self._queueRedraw(which='targets')

        elif changedTargetAttrs == ['isSelected']:
            pass  # selection update will be handled separately

        else:
            # assume anything/everything changed, clear target and start over
            self._redrawTargetKeys.update(changedTargetKeys)
            self._queueRedraw(which='targets')

    def _redraw(self, which: tp.Union[str | None, list[str]] = None, **kwargs):
        super()._redraw(which=which)

        if which is None:
            self._redraw(which='all', **kwargs)
            return

        if not isinstance(which, str):
            for subWhich in which:
                self._redraw(which=subWhich, **kwargs)
            return

        if which == 'all':
            self._redraw(which=['targets'])
            return

        if which == 'targets':
            changedTargetKeys = self._redrawTargetKeys.copy()
            self._redrawTargetKeys.clear()
            if self._hasInitialized or self._isInitializing:
                # update views
                for viewKey, view in self._views.items():
                    for key in changedTargetKeys:
                        try:
                            target = self.session.targets[key]
                        except KeyError as e:
                            # previous key no longer in targets
                            if viewKey + key in self._targetActors:
                                visualizedTarget = self._targetActors.pop(viewKey + key)
                                visualizedTarget.clearActors()
                        else:
                            if target.isVisible:
                                self._createVisualForTarget(viewKey=viewKey, target=target)

                    view.updateView()

                currentTarget = self._getCurrentTargetKey()
                self._onCurrentTargetChanged(currentTarget)

    def _onImportTargetsBtnClicked(self, checked: bool):
        newFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(self._wdgt,
                                                               'Select targets file to import',
                                                               os.path.dirname(self.session.filepath),
                                                               'json (*.json);; NaviNIBS (*.navinibs)')

        if len(newFilepath) == 0:
            logger.warning('Import cancelled')
            return

        self._importTargetsFromFile(newFilepath=newFilepath)

    def _importTargetsFromFile(self, newFilepath: str):

        self.session.mergeFromFile(filepath=newFilepath, sections=['targets'])

        self._tableWdgt.resizeColumnsToContents()  # resize immediately rather than waiting for delayed auto-resize

    def _onAddBtnClicked(self, checked: bool):
        # create new target at current crosshair location, and autoset entry
        targetCoord = self._getCrosshairCoord()
        targetKey = makeStrUnique('Target-1', existingStrs=self.session.targets.keys(), delimiter='-')
        currentTargetKey = self._tableWdgt.currentCollectionItemKey
        if currentTargetKey is not None:
            # match angle from midline of current target for target in new position
            nextTargetAngle = self.session.targets[currentTargetKey].calculatedAngle
        else:
            nextTargetAngle = 0
        target = Target(targetCoord=targetCoord,
                        angle=nextTargetAngle,
                        key=targetKey,
                        session=self.session)
        target.autosetEntryCoord()
        self.session.targets.addItem(target)
        self._tableWdgt.currentCollectionItemKey = targetKey  # select new target

    def _onDeleteBtnClicked(self, checked: bool):
        selectedTargetKeys = self._tableWdgt.selectedCollectionItemKeys
        self.session.targets.deleteItems(selectedTargetKeys)

    def _onDuplicateBtnClicked(self, checked: bool):
        currentTargetKey = self._tableWdgt.currentCollectionItemKey
        assert currentTargetKey is not None
        if '_' in currentTargetKey:
            delim = '_'
        elif ' ' in currentTargetKey:
            delim = ' '
        else:
            delim = '-'
        newTargetKey = makeStrUnique(f'{currentTargetKey}{delim}copy{delim}1', existingStrs=self.session.targets.keys(), delimiter=delim)
        logger.info(f'Duplicating target {currentTargetKey} to {newTargetKey}')
        currentTarget = self.session.targets[currentTargetKey]
        newTarget = currentTarget.asDict()
        newTarget['key'] = newTargetKey
        self.session.targets.addItem(Target.fromDict(newTarget))

    def _onGotoBtnClicked(self, checked: bool):
        targetKey = self._getCurrentTargetKey()
        if targetKey is not None:
            self._gotoTarget(self._getCurrentTargetKey())
        else:
            logger.warning('No target selected')














from __future__ import annotations

import attrs
import copy
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import shutil
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish, attrsWithNumpyAsDict, attrsWithNumpyFromDict

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem


logger = logging.getLogger(__name__)


@attrs.define
class DigitizedLocation(GenericCollectionDictItem[str]):
    _plannedCoord: tp.Optional[np.ndarray] = None
    _plannedCoordSpace: tp.Optional[str] = None  # str key referring to a CoordinateSystem (e.g. MNI)
    _sampledCoord: tp.Optional[np.ndarray] = None  # in MRI space
    _type: tp.Optional[str] = None

    _color: str = '#6f2da8'

    @property
    def plannedCoord(self):
        return self._plannedCoord

    @property
    def plannedCoordSpace(self):
        return self._plannedCoordSpace

    @property
    def sampledCoord(self):
        return self._sampledCoord

    @sampledCoord.setter
    def sampledCoord(self, newCoord: tp.Optional[np.ndarray]):
        if array_equalish(newCoord, self._sampledCoord):
            return
        self.sigItemAboutToChange.emit(self.key, ['sampledCoord'])
        self._sampledCoord = newCoord
        self.sigItemChanged.emit(self.key, ['sampledCoord'])

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, newType: tp.Optional[str]):
        if newType == self._type:
            return
        self.sigItemAboutToChange.emit(self.key, ['type'])
        self._type = newType
        self.sigItemChanged.emit(self.key, ['type'])

    @property
    def color(self):
        return self._color

    def asDict(self) -> dict[str, tp.Any]:
        return attrsWithNumpyAsDict(self, npFields=('plannedCoord', 'sampledCoord'))

    @classmethod
    def fromDict(cls, d: dict[str, tp.Any]):
        return attrsWithNumpyFromDict(cls, d, npFields=('plannedCoord', 'sampledCoord'))


@attrs.define
class DigitizedLocations(GenericCollection[str, DigitizedLocation]):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def loadFromXML(self, xmlPath: str):
        import xmltodict
        with open(xmlPath, 'r') as f:
            montage = xmltodict.parse(f.read())

        for marker in montage['PresetEEG']['Marker']:
            self.addItem(DigitizedLocation(
                key=marker['@name'],
                color=marker['@color']
            ))

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> DigitizedLocations:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = DigitizedLocation.fromDict(itemDict)

        return cls(items=items)


from __future__ import annotations

import asyncio
import os.path

import attrs
import logging
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.Model.Session import Session, Tool, Fiducial, Target, Sample, DigitizedLocation, TriggerSource, Addon
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


@attrs.define
class SessionTreeItem:
    _label: str
    _obj: object = None
    _parent: SessionTreeItem | None = None
    _children: list[SessionTreeItem] = attrs.field(factory=list, init=False)
    _checked: QtCore.Qt.CheckState | None = attrs.field(default=None, init=False)

    sigCheckedChanged: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        if self._parent is not None:
            self._parent.addChild(self)

    def addChild(self, child: 'SessionTreeItem'):
        if child in self._children:
            return
        assert child._parent is None or child._parent is self, "Child already has a parent"
        child._parent = self
        self._children.append(child)
        child.sigCheckedChanged.connect(self._onChildCheckedChanged)

    def child(self, row):
        return self._children[row]

    def childByLabel(self, label: str) -> SessionTreeItem | None:
        for child in self._children:
            if child.label == label:
                return child
        return None

    def row(self):
        if self._parent:
            return self._parent._children.index(self)
        return 0
    
    def _onChildCheckedChanged(self):
        if all(child.checked == QtCore.Qt.Checked for child in self._children):
            self.checked = QtCore.Qt.Checked
        elif all(child.checked == QtCore.Qt.Unchecked for child in self._children):
            self.checked = QtCore.Qt.Unchecked
        else:
            self.checked = QtCore.Qt.PartiallyChecked

    @property
    def label(self):
        return self._label

    @property
    def obj(self):
        return self._obj

    @property
    def children(self):
        return self._children

    @property
    def childrenLabels(self) -> list[str]:
        return [child.label for child in self._children]

    @property
    def parent(self):
        return self._parent

    @property
    def checked(self):
        if self._checked is None:
            if len(self._children) > 0:
                # If there are children, check their state
                if all(child.checked == QtCore.Qt.Checked for child in self._children):
                    self.checked = QtCore.Qt.Checked
                elif all(child.checked == QtCore.Qt.Unchecked for child in self._children):
                    self.checked = QtCore.Qt.Unchecked
                else:
                    self.checked = QtCore.Qt.PartiallyChecked
            else:
                # If there are no children, default to unchecked
                self.checked = QtCore.Qt.Unchecked
        
        return self._checked

    @checked.setter
    def checked(self, value: QtCore.Qt.CheckState):
        if value == self._checked:
            return
        logger.debug(f'Setting checked state of {self._label} to {value}')
        self._checked = value
        if self._checked != QtCore.Qt.PartiallyChecked:
            # Update the checked state of all children
            for child in self._children:
                with child.sigCheckedChanged.disconnected(self._onChildCheckedChanged):
                    child.checked = value
        self.sigCheckedChanged.emit()


class SessionTreeModel(QtCore.QAbstractItemModel):
    """Model for displaying session elements in a tree view."""

    def __init__(self, session: Session, parent=None):
        super().__init__(parent)

        self.sigItemCheckedChanged = Signal((SessionTreeItem,))

        self._session = session
        self._rootItem = SessionTreeItem('Session', session)
        self._selectionModel = None  # Will be set by the view

        if self._session.subjectID is not None:
            item = SessionTreeItem(label='Subject ID', obj=self._session.subjectID, parent=self._rootItem)

        if self._session.sessionID is not None:
            item = SessionTreeItem(label='Session ID', obj=self._session.sessionID, parent=self._rootItem)

        if len(self._session.MRI.asDict(filepathRelTo=session.filepath)) > 0:
            item = SessionTreeItem(label='MRI', obj=self._session.MRI, parent=self._rootItem)

        if len(self._session.headModel.asDict(filepathRelTo=session.filepath)) > 0:
            item = SessionTreeItem(label='Head Model', obj=self._session.headModel, parent=self._rootItem)

        if len(self._session.coordinateSystems.asList()) > 0:  # use .asList to ignore autogenerated coordinate systems
            item = SessionTreeItem(label='Coordinate Systems', obj=None, parent=self._rootItem)
            for key, cs in self._session.coordinateSystems.items():
                child = SessionTreeItem(label=key, obj=cs, parent=item)

        if len(self._session.subjectRegistration.asDict()) > 0:
            item = SessionTreeItem(label='Subject Registration', obj=None, parent=self._rootItem)
            if len(self._session.subjectRegistration.fiducials) > 0:
                # separate planned and sampled fiducials
                subitem = SessionTreeItem(label='Planned fiducials', obj=None, parent=item)
                for key, fid in self._session.subjectRegistration.fiducials.items():
                    if fid.plannedCoord is not None:
                        child = SessionTreeItem(label=key, obj=fid, parent=subitem)
                subitem = SessionTreeItem(label='Sampled fiducials', obj=None, parent=item)
                for key, fid in self._session.subjectRegistration.fiducials.items():
                    if fid.sampledCoord is not None:
                        child = SessionTreeItem(label=key, obj=fid, parent=subitem)

            if len(self._session.subjectRegistration.sampledHeadPoints) > 0:
                subitem = SessionTreeItem(label='Sampled head points', obj=None, parent=item)

            if self._session.subjectRegistration.trackerToMRITransf is not None:
                subitem = SessionTreeItem(label='Tracker to MRI transf', obj=None, parent=item)

        if len(self._session.targets.asList()) > 0:
            item = SessionTreeItem(label='Targets', obj=None, parent=self._rootItem)
            for target in self._session.targets.values():
                child = SessionTreeItem(label=target.key, obj=target, parent=item)

        toolList = self._session.tools.asList()
        if len(toolList) > 0:
            item = SessionTreeItem(label='Tools', obj=None, parent=self._rootItem)
            for toolDict in toolList:
                child = SessionTreeItem(label=toolDict['key'], obj=toolDict, parent=item)

        if len(self._session.samples) > 0:
            item = SessionTreeItem(label='Samples', obj=None, parent=self._rootItem)
            for key, sample in self._session.samples.items():
                child = SessionTreeItem(label=key, obj=sample, parent=item)

        if len(self._session.digitizedLocations) > 0:
            item = SessionTreeItem(label='Digitized Locations', obj=None, parent=self._rootItem)
            # separate planned and sampled information
            key = '<Planned>'
            subitem = SessionTreeItem(label=key, obj=None, parent=item)
            for key, loc in self._session.digitizedLocations.items():
                child = SessionTreeItem(label=key, obj=loc, parent=subitem)
            if any(loc.sampledCoord is not None for loc in self._session.digitizedLocations.values()):
                key = '<Sampled>'
                subitem = SessionTreeItem(label=key, obj=None, parent=item)
                for key, loc in self._session.digitizedLocations.items():
                    if loc.sampledCoord is not None:
                        child = SessionTreeItem(label=key, obj=loc, parent=subitem)

        if len(self._session.triggerSources) > 0:
            item = SessionTreeItem(label='Trigger Sources', obj=None, parent=self._rootItem)
            for key, ts in self._session.triggerSources.items():
                child = SessionTreeItem(label=key, obj=ts, parent=item)

        if len(self._session.dockWidgetLayouts) > 0:
            item = SessionTreeItem(label='Dock Widget Layouts', obj=None, parent=self._rootItem)

        if len(self._session.addons) > 0:
            item = SessionTreeItem(label='Addons', obj=None, parent=self._rootItem)
            for key, addon in self._session.addons.items():
                subitem = SessionTreeItem(label=key, obj=None, parent=item)
                child = SessionTreeItem(label='Addon install path', obj=addon, parent=subitem)
                if len(set(addon.asDict().keys()) - {'addonInstallPath', 'addonVersion'}) > 0:
                    # TODO: add support for addons to declare more specific subitems
                    child = SessionTreeItem(label='Addon settings', obj=None, parent=subitem)

        def connectCheckedSignal(item: SessionTreeItem):
            item.sigCheckedChanged.connect(lambda item=item: self._onItemCheckedChanged(item))
            for child in item.children:
                connectCheckedSignal(child)

        connectCheckedSignal(self._rootItem)

    def setSelectionModel(self, selectionModel):
        self._selectionModel = selectionModel

    def index(self, row, column, parent=QtCore.QModelIndex()):
        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        if not parent.isValid():
            # No parent, so we are at the top level
            if row < 1:
                return self.createIndex(row, column, self._rootItem)
        else:
            # Child items
            parentItem: SessionTreeItem = parent.internalPointer()
            if row < len(parentItem.children):
                return self.createIndex(row, column, parentItem.children[row])

        return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()

        item = index.internalPointer()
        if item.parent is None:
            return QtCore.QModelIndex()
        return self._indexForItem(item.parent)
        
    def _indexForItem(self, item: SessionTreeItem):
        """Get the index for a given item"""
        if item is self._rootItem:
            return self.createIndex(0, 0, item)
        assert item.parent is not None
        row = item.parent.children.index(item)
        return self.createIndex(row, 0, item)

    def rowCount(self, parent=QtCore.QModelIndex()):
        if not parent.isValid():
            return 1
        item = parent.internalPointer()
        return len(item.children)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 1

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return None
        item = index.internalPointer()
        if role == QtCore.Qt.DisplayRole:
            return item.label
        elif role == QtCore.Qt.CheckStateRole:
            return item.checked
        return None

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsUserCheckable

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            item = index.internalPointer()
            value = QtCore.Qt.CheckState(value)  # convert from generic value to CheckState
            # If selection model is set and item is selected, apply to all selected items
            if self._selectionModel is not None and self._selectionModel.isSelected(index):
                for selected_index in self._selectionModel.selectedIndexes():
                    selected_item = selected_index.internalPointer()
                    selected_item.checked = value
                    self.dataChanged.emit(selected_index, selected_index, [QtCore.Qt.CheckStateRole])
                return True
            else:
                item.checked = value
                return True
        return False

    def _onItemCheckedChanged(self, item: SessionTreeItem):
        """Handle the checked state change of an item"""
        index = self._indexForItem(item)
        assert index.isValid(), "Item index is not valid"
        self.dataChanged.emit(
            index, index, [QtCore.Qt.CheckStateRole])
        self.sigItemCheckedChanged.emit(item)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        # if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
        #     return "Session Elements"
        return None

    @property
    def rootItem(self):
        return self._rootItem


@attrs.define
class ImportSessionWindow:
    _parent: QtWidgets.QWidget
    _session: Session = attrs.field(repr=False)
    _otherSession: Session = attrs.field(repr=False)

    _wdgt: QtWidgets.QDialog = attrs.field(init=False)
    _presetsComboBox: QtWidgets.QComboBox = attrs.field(init=False)
    _sessionTreeView: QtWidgets.QTreeView = attrs.field(init=False)
    _sessionTreeModel: SessionTreeModel = attrs.field(init=False)
    _finalizeButtonBox: QtWidgets.QDialogButtonBox = attrs.field(init=False)

    _presetComboBoxNeedsRefresh: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    sigFinished: Signal = attrs.field(init=False, factory=lambda: Signal((bool,)))

    def __attrs_post_init__(self):
        self._wdgt = QtWidgets.QDialog(self._parent)
        self._wdgt.setModal(True)

        self._wdgt.setWindowTitle('Import session')

        self._wdgt.setWindowModality(QtGui.Qt.WindowModal)

        self._wdgt.finished.connect(self._onDlgFinished)

        topLayout = QtWidgets.QVBoxLayout()
        self._wdgt.setLayout(topLayout)

        formContainer = QtWidgets.QWidget()
        topLayout.addWidget(formContainer)
        formLayout = QtWidgets.QFormLayout()
        formContainer.setLayout(formLayout)
        self._presetsComboBox = QtWidgets.QComboBox(self._wdgt)
        self._presetsComboBox.addItems(self._getPresetKeys())
        self._presetsComboBox.currentIndexChanged.connect(self._onPresetGUIChanged)
        formLayout.addRow('Preset', self._presetsComboBox)

        # Add session tree view
        treeContainer = QtWidgets.QWidget()
        topLayout.addWidget(treeContainer)
        treeLayout = QtWidgets.QVBoxLayout()
        treeContainer.setLayout(treeLayout)

        treeLabel = QtWidgets.QLabel("Select elements to import:")
        treeLayout.addWidget(treeLabel)

        self._sessionTreeView = QtWidgets.QTreeView(self._wdgt)
        # Use ExtendedSelection mode to support shift-selection
        self._sessionTreeView.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self._sessionTreeView.setHeaderHidden(True)
        self._sessionTreeModel = SessionTreeModel(self._otherSession, self._wdgt)
        self._sessionTreeModel.sigItemCheckedChanged.connect(self._onTreeModelItemCheckedChanged)
        self._sessionTreeView.setModel(self._sessionTreeModel)
        self._sessionTreeView.setMinimumHeight(300)
        self._sessionTreeView.setMinimumWidth(300)
        self._sessionTreeView.expandToDepth(0)
        self._sessionTreeModel.setSelectionModel(self._sessionTreeView.selectionModel())

        treeLayout.addWidget(self._sessionTreeView)

        # set a reasonable default for preset selection
        if self._session.subjectID is not None and self._session.subjectID == self._otherSession.subjectID:
            self._presetsComboBox.setCurrentText('Same subject, different session')
        else:
            self._presetsComboBox.setCurrentText('Different subject')

        # Add button box
        self._finalizeButtonBox = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self._finalizeButtonBox.accepted.connect(self._onAccept)
        self._finalizeButtonBox.rejected.connect(self._wdgt.reject)
        topLayout.addWidget(self._finalizeButtonBox)

        # TODO: autoselect default preset based on whether session subject IDs match

        self._wdgt.show()

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_refreshComboBox))

    def _getPresetKeys(self) -> list[str]:
        return ['Custom', 'Different subject', 'Same subject, different session']

    def _onTreeModelItemCheckedChanged(self, item: SessionTreeItem):
        self._presetComboBoxNeedsRefresh.set()

    async def _loop_refreshComboBox(self):
        while True:
            await self._presetComboBoxNeedsRefresh.wait()
            if self._presetComboBoxNeedsRefresh.is_set():
                self._presetComboBoxNeedsRefresh.clear()
                self._presetsComboBox.setCurrentText('Custom')

    def _onPresetGUIChanged(self, index: int):
        preset = self._getPresetKeys()[index]

        if preset == 'Custom':
            # don't change any selections
            return

        # Deselect and uncheck all items
        self._sessionTreeView.clearSelection()
        self._uncheckAllItems()

        # Helper to check by label
        def check_by_label(label):
            idx = self._findCategoryIndexByLabel(label)
            if idx is not None:
                self._checkCategoryByIndex(idx)

        # Select and check items based on preset
        if preset == 'Different subject':
            # Only select items that don't depend on subject
            check_by_label('Tools')
            check_by_label(('Digitized Locations', '<Planned>'))
            check_by_label('Trigger Sources')
            # check_by_label('Dock Widget Layouts')  # TODO: uncomment
            # TODO: make active dock widget the session info tab (rather than for example jumping straight to navigate in a new subject)
            check_by_label('Addons')
            # TODO: add support for addons to declare specific fields as subject-specific or not
        elif preset == 'Same subject, different session':
            # Select subject-specific items
            check_by_label('Subject ID')
            check_by_label('MRI')
            check_by_label('Head Model')
            check_by_label('Coordinate Systems')
            check_by_label(('Subject Registration', 'Planned fiducials'))
            check_by_label('Targets')
            check_by_label('Tools')
            check_by_label(('Digitized Locations', '<Planned>'))
            check_by_label('Trigger Sources')
            # check_by_label('Dock Widget Layouts')  # TODO: uncomment
            # TODO: make active dock widget the session info tab (rather than for example jumping straight to navigate in a new subject)
            check_by_label('Addons')
            # TODO: add support for addons to declare specific fields as session-specific or not

        self._presetComboBoxNeedsRefresh.clear()  # don't change to custom in response to recent changes

    def _findCategoryIndexByLabel(self, label: str | tuple[str,...], parent: QtCore.QModelIndex = None) -> int | None:
        """Return the index of the top-level category with the given label, or None if not found."""
        if isinstance(label, str):
            return self._findCategoryIndexByLabel(label=(label,), parent=parent)

        if parent is None:
            parent = self._sessionTreeModel.index(0, 0)

        for i in range(self._sessionTreeModel.rowCount(parent=parent)):
            model_index = self._sessionTreeModel.index(i, 0, parent)
            item = model_index.internalPointer()
            if item and getattr(item, 'label', None) == label[0]:
                if len(label) > 1:
                    # If there are more labels to check, recurse into the children
                    return self._findCategoryIndexByLabel(label[1:], model_index)
                # If this is the last label, return the index
                return model_index
        logger.warning(f"Category with label '{label}' not found in the session tree.")
        return None

    def _uncheckAllItems(self):
        """Uncheck all items in the tree"""
        for i in range(self._sessionTreeModel.rowCount()):
            model_index = self._sessionTreeModel.index(i, 0)
            self._sessionTreeModel.setData(model_index, QtCore.Qt.Unchecked, QtCore.Qt.CheckStateRole)

    def _checkCategoryByIndex(self, index: QtCore.QModelIndex):
        """Check a top-level category by index"""
        if False:
            # also select
            self._sessionTreeView.selectionModel().select(
                index, QtCore.QItemSelectionModel.Select)
        self._sessionTreeModel.setData(index, QtCore.Qt.Checked, QtCore.Qt.CheckStateRole)

    def _selectCategoryByIndex(self, index: QtCore.QModelIndex):
        """Select a top-level category by index"""
        self._sessionTreeView.selectionModel().select(
            index, QtCore.QItemSelectionModel.Select)

    def _onDlgFinished(self, result: int):
        self.sigFinished.emit(result == QtWidgets.QDialog.Accepted)

    def _onAccept(self):
        """Handle import when OK is clicked"""

        # TODO: block at least some session update signals until after all changes are complete
        # (e.g. to not have a view start using new MRI data before a new head model is also updated)

        item = self._sessionTreeModel.rootItem.childByLabel('Subject ID')
        if item is not None and item.checked == QtCore.Qt.CheckState.Checked:
            self.session.subjectID = self.otherSession.subjectID

        item = self._sessionTreeModel.rootItem.childByLabel('Session ID')
        if item is not None and item.checked == QtCore.Qt.CheckState.Checked:
            self.session.sessionID = self.otherSession.sessionID

        item = self._sessionTreeModel.rootItem.childByLabel('MRI')
        if item is not None and item.checked == QtCore.Qt.CheckState.Checked:
            kwargs = self.otherSession.MRI.asDict(filepathRelTo=self.otherSession.filepath)
            if 'filepath' in kwargs:
                # TODO: if other session referenced a file inside its own folder, copy to primary session folder here
                # or if relative path is different between sessions (e.g. due to them being in different parent directories)
                # offer to copy the file here
                kwargs['filepath'] = os.path.join(self.otherSession.filepath, kwargs['filepath'])

            for key, val in kwargs.items():
                setattr(self.session.MRI, key, val)

        item = self._sessionTreeModel.rootItem.childByLabel('Head Model')
        if item is not None and item.checked == QtCore.Qt.CheckState.Checked:
            kwargs = self.otherSession.headModel.asDict(filepathRelTo=self.otherSession.filepath)
            if 'filepath' in kwargs:
                # TODO: if other session referenced a file inside its own folder, copy to primary session folder here
                # or if relative path is different between sessions (e.g. due to them being in different parent directories)
                # offer to copy the file here
                # note: if copying files here, need to copy entire SimNIBS folder, not just the referenced .msh file
                kwargs['filepath'] = os.path.join(self.otherSession.filepath, kwargs['filepath'])
            for key, val in kwargs.items():
                setattr(self.session.headModel, key, val)

        item = self._sessionTreeModel.rootItem.childByLabel('Coordinate Systems')
        if item is not None:
            for subitem in item.children:
                if subitem.checked != QtCore.Qt.CheckState.Checked:
                    continue
                coordSys = self.otherSession.coordinateSystems[subitem.label]
                # TODO: maybe convert from/to dict as intermediate rather than directly copying ref to same instance of coordSys between sessions
                self.session.coordinateSystems[coordSys.key] = coordSys

        item = self._sessionTreeModel.rootItem.childByLabel('Subject Registration')
        if item is not None:
            copyPlannedFiducials: list[str] = []
            subitem = item.childByLabel('Planned fiducials')
            if subitem is not None:
                for subsubitem in subitem.children:
                    if subsubitem.checked != QtCore.Qt.CheckState.Checked:
                        continue
                    copyPlannedFiducials.append(subsubitem.label)

            copySampledFiducials: list[str] = []
            subitem = item.childByLabel('Sampled fiducials')
            if subitem is not None:
                for subsubitem in subitem.children:
                    if subsubitem.checked != QtCore.Qt.CheckState.Checked:
                        continue
                    copySampledFiducials.append(subsubitem.label)

            for fid in self.otherSession.subjectRegistration.fiducials.values():
                fidDict = fid.asDict()
                doCopy = False
                if fid.key not in copyPlannedFiducials:
                    fidDict.pop('plannedCoord', None)
                else:
                    doCopy = True

                if fid.key not in copySampledFiducials:
                    fidDict.pop('sampledCoord', None)
                    fidDict.pop('sampledCoords', None)
                else:
                    doCopy = True

                if doCopy:
                    if fid.key in self.session.subjectRegistration.fiducials:
                        for key, val in fidDict.items():
                            setattr(self.session.subjectRegistration.fiducials[fid.key], key, val)
                    else:
                        self.session.subjectRegistration.fiducials[fid.key] = Fiducial.fromDict(fidDict)

            subitem = item.childByLabel('Sampled head points')
            if subitem is not None and subitem.checked == QtCore.Qt.CheckState.Checked:
                self.session.subjectRegistration.sampledHeadPoints.replace(self.otherSession.subjectRegistration.sampledHeadPoints.asNDArray())
                self.session.subjectRegistration.sampledHeadPoints.alignmentWeights = self.otherSession.subjectRegistration.sampledHeadPoints.alignmentWeights

            subitem = item.childByLabel('Tracker to MRI transf')
            if subitem is not None and subitem.checked == QtCore.Qt.CheckState.Checked:
                self.session.subjectRegistration.trackerToMRITransf = self.otherSession.subjectRegistration.trackerToMRITransf

        item = self._sessionTreeModel.rootItem.childByLabel('Targets')
        targetsTreeItem = item
        if item is not None:
            for subitem in item.children:
                if subitem.checked != QtCore.Qt.CheckState.Checked:
                    continue
                target = self.otherSession.targets[subitem.label]
                self.session.targets[target.key] = Target.fromDict(target.asDict(), session=self.session)

        item = self._sessionTreeModel.rootItem.childByLabel('Tools')
        if item is not None:
            for subitem in item.children:
                if subitem.checked != QtCore.Qt.CheckState.Checked:
                    continue
                if subitem.label == 'ToolPositionsServer':
                    # handle special ToolPositionsServer info that may be listed next to actual tools
                    kwargs = self.otherSession.tools.positionsServerInfo.asDict()
                    for key, val in kwargs.items():
                        setattr(self.session.tools.positionsServerInfo, key, val)
                else:
                    tool = self.otherSession.tools[subitem.label]
                    self.session.tools[tool.key] = self.session.tools.toolFromDict(tool.asDict())

        item = self._sessionTreeModel.rootItem.childByLabel('Samples')
        if item is not None:
            for subitem in item.children:
                if subitem.checked != QtCore.Qt.CheckState.Checked:
                    continue
                sample = self.otherSession.samples[subitem.label]
                if sample.targetKey is not None:
                    # make sure corresponding target was also imported/copied
                    if targetsTreeItem is None or \
                            (targetItem := targetsTreeItem.childByLabel(sample.targetKey)) is not None or \
                            targetItem.checked != QtCore.Qt.CheckState.Checked:
                        logger.error(f'Target {sample.targeKey} not selected for import, but referenced by sample {sample.key}')
                        raise KeyError('When importing samples, corresponding target(s) must also be imported')
                self.session.samples[sample.key] = Sample.fromDict(sample.asDict())

        item = self._sessionTreeModel.rootItem.childByLabel('Digitized Locations')
        if item is not None:
            copyPlannedLocations: list[str] = []
            subitem = item.childByLabel('<Planned>')
            if subitem is not None:
                for subsubitem in subitem.children:
                    if subsubitem.checked != QtCore.Qt.CheckState.Checked:
                        continue
                    copyPlannedLocations.append(subsubitem.label)

            copySampledLocations: list[str] = []
            subitem = item.childByLabel('<Sampled>')
            if subitem is not None:
                for subsubitem in subitem.children:
                    if subsubitem.checked != QtCore.Qt.CheckState.Checked:
                        continue
                    copySampledLocations.append(subsubitem.label)

            for loc in self.otherSession.digitizedLocations.values():
                locDict = loc.asDict()
                doCopy = False
                if loc.key not in copyPlannedLocations:
                    locDict.pop('plannedCoord', None)
                else:
                    doCopy = True

                if loc.key not in copySampledLocations:
                    locDict.pop('sampledCoord', None)
                else:
                    doCopy = True

                if doCopy:
                    if loc.key in self.session.digitizedLocations:
                        for key, val in locDict.items():
                            setattr(self.session.digitizedLocations[loc.key], key, val)
                    else:
                        self.session.digitizedLocations[loc.key] = DigitizedLocation.fromDict(locDict)

        item = self._sessionTreeModel.rootItem.childByLabel('Trigger Sources')
        if item is not None:
            for subitem in item.children:
                if subitem.checked != QtCore.Qt.CheckState.Checked:
                    continue
                trigSource = self.otherSession.triggerSources[subitem.label]
                self.session.triggerSources[trigSource.key] = TriggerSource.fromDict(trigSource.asDict())

        item = self._sessionTreeModel.rootItem.childByLabel('Dock Widget Layouts')
        if item is not None and item.checked == QtCore.Qt.CheckState.Checked:
            raise NotImplementedError  # TODO: add support for restoring layout(s)

        item = self._sessionTreeModel.rootItem.childByLabel('Addons')
        if item is not None:
            for subitem in item.children:
                if subitem.checked == QtCore.Qt.CheckState.Unchecked:
                    continue
                addon = self.otherSession.addons[subitem.label]

                if addon.key in self.session.addons:
                    # make sure install path is the same for both addons
                    existingAddon = self.session.addons[addon.key]
                    if not os.path.abspath(addon.addonInstallPath) == os.path.abspath(existingAddon.addonInstallPath):
                        raise RuntimeError('Replacing existing addon with another version of same addon not supported. Make edits manually in JSON file(s) and restart NaviNIBS instead.')

                    raise NotImplementedError  # TODO: copy over sessionAttrs

                else:
                    addonDict = addon.asDict()
                    self.session.addons[addon.key] = Addon.fromDict(addonDict)

        self._wdgt.accept()

    def show(self):
        self._wdgt.show()

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def session(self):
        return self._session

    @property
    def otherSession(self):
        return self._otherSession





from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView
from NaviNIBS.util.pyvista import Actor
from NaviNIBS.util.pyvista import (
    DefaultBackgroundPlotter,
    DefaultPrimaryLayeredPlotter,
    DefaultSecondaryLayeredPlotter,
    RemotePlotterProxy
)
from NaviNIBS.util.pyvista.PlotInteraction import set_mouse_event_for_picking
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import composeTransform, applyTransform
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.Navigator.Model.Session import Session


logger = logging.getLogger(__name__)


@attrs.define()
class SurfSliceView(MRISliceView):

    _activeSurf: tp.Union[str, tp.List[str]] = 'gmSurf'

    _slicePlotMethod: str = 'cameraClippedVolume'

    _doLayeredPlotters: bool = True
    _primaryPlotter: DefaultPrimaryLayeredPlotter = attrs.field(init=False, default=None)
    _plotter: DefaultSecondaryLayeredPlotter | None = attrs.field(init=False, default=None)  # replace parent classes' plotter with layered plotter
    _surfPlotter: DefaultSecondaryLayeredPlotter = attrs.field(init=False)

    _opacity: float = 0.5

    _surfColor: tp.Union[str, tp.List[str]] = '#d9a5b2'
    _surfOpacity: tp.Union[float, tp.List[float]] = 0.5
    _surfPlotInitialized: bool = attrs.field(init=False, default=False)
    _surfPlotActors: list[Actor] = attrs.field(init=False, factory=list)

    def __attrs_post_init__(self):

        if self._primaryPlotter is None:
            if self._doLayeredPlotters:
                self._primaryPlotter = DefaultPrimaryLayeredPlotter(
                    doAutoAdjustCameraClippingRange=False,
                )
                self._plotter = self._primaryPlotter  # will be replaced with secondary plot layer during async init
            else:
                self._primaryPlotter = DefaultBackgroundPlotter()
                self._plotter = self._primaryPlotter


        else:
            pass  # presumably initialized by subclass

        super().__attrs_post_init__()

        if self._session is not None:
            self._session.headModel.sigDataChanged.connect(self._onHeadModelDataChanged)

    async def _finish_init(self):
        if isinstance(self._primaryPlotter, RemotePlotterProxy):
            await self._primaryPlotter.isReadyEvent.wait()

        if self._doLayeredPlotters:
            self._plotter = self._primaryPlotter.addLayeredPlotter(key='MRISlice', layer=0)
            self._surfPlotter = self._primaryPlotter.addLayeredPlotter(key='SurfSlice', layer=1)
        else:
            self._surfPlotter = self._primaryPlotter

        await super()._finish_init()

        with self._primaryPlotter.allowNonblockingCalls():
            self._primaryPlotter.set_background(self._backgroundColor)

        with self._surfPlotter.allowNonblockingCalls():
            self._surfPlotter.enable_depth_peeling(8)

    @MRISliceView.wdgt.getter
    def wdgt(self) -> QtWidgets.QWidget:
        return self._primaryPlotter

    @MRISliceView.plotter.getter
    def plotter(self) -> DefaultPrimaryLayeredPlotter:
        return self._primaryPlotter

    @property
    def activeSurf(self):
        return self._activeSurf

    @activeSurf.setter
    def activeSurf(self, newKey: str):
        if self._activeSurf == newKey:
            return
        self._clearSurfPlotActors()
        self._activeSurf = newKey
        self._onHeadModelDataChanged(whatChanged=self._activeSurf)

    def _clearPlot(self):
        self._clearSurfPlotActors()
        super()._clearPlot()

    def _clearSurfPlotActors(self):
        with self._surfPlotter.allowNonblockingCalls():
            for actor in self._surfPlotActors:
                self._surfPlotter.remove_actor(actor)
        self._surfPlotActors.clear()
        self._surfPlotInitialized = False

    def _onHeadModelDataChanged(self, whatChanged: str):
        if whatChanged == self._activeSurf:
            self._clearSurfPlotActors()
            self.updateView()
        else:
            # ignore other changes
            pass

    def _updateView(self):

        super()._updateView()

        if not self.finishedAsyncInit.is_set():
            # plotter not available yet
            return

        if isinstance(self._activeSurf, str):
            surfKeys = [self._activeSurf]
        else:
            surfKeys = self._activeSurf

        if isinstance(self._surfColor, str):
            surfColors = [self._surfColor]
        else:
            surfColors = self._surfColor

        if not isinstance(self._surfOpacity, list):
            surfOpacities = [self._surfOpacity]
        else:
            surfOpacities = self._surfOpacity

        if not self._surfPlotInitialized \
                and self.session is not None:
            actors = []
            for iSurf, surfKey in enumerate(surfKeys):
                if getattr(self.session.headModel, self._activeSurf) is not None:
                    actor = self._surfPlotter.add_mesh(mesh=getattr(self.session.headModel, surfKey),
                                                                 color=surfColors[iSurf % len(surfColors)],
                                                                 opacity=surfOpacities[iSurf % len(surfOpacities)],
                                                                 name=self.label + '_' + surfKey + '_surf',
                                                                 render=False,
                                                                 reset_camera=False
                                                                 )
                    actors.append(actor)
                    self._surfPlotInitialized = True

            self._surfPlotActors.extend(actors)

        if self._slicePlotMethod == 'cameraClippedVolume':
            with self._primaryPlotter.allowNonblockingCalls():
                self._primaryPlotter.set_camera_clipping_range((self._cameraOffsetDist * 0.98, self._cameraOffsetDist * 1.02))

        with self.plotter.allowNonblockingCalls():
            self.plotter.render()


@attrs.define()
class Surf3DView(SurfSliceView):
    _surfOpacity: float = 1.

    _doLayeredPlotters: bool = False

    _pickableSurfs: list[str] | None = None

    def __attrs_post_init__(self):
        if self._doLayeredPlotters:
            self._primaryPlotter = DefaultPrimaryLayeredPlotter(
                doAutoAdjustCameraClippingRange=True,
            )
        else:
            pass  # let superclass initialize plotter

        super().__attrs_post_init__()

    def _updateView(self):

        if not self.finishedAsyncInit.is_set():
            # plotter not available yet
            return

        if self.session is None or self.session.MRI.data is None:
            # no data available
            if self._plotterInitialized:
                logger.debug('Clearing plot for {} slice'.format(self.label))
                with self._plotter.allowNonblockingCalls():
                    self._plotter.clear()
                self._surfPlotter.clear()

                self.sliceOrigin = None
                self._plotterInitialized = False
            return

        # data available, update display
        logger.debug('Updating plot for {} slice'.format(self.label))
        if self._sliceOrigin is None:
            self.sliceOrigin = (self.session.MRI.data.affine @ np.append(np.asarray(self.session.MRI.data.shape) / 2,
                                                                         1))[:-1]
            return  # prev line will have triggered its own update

        if not self._surfPlotInitialized and self.session is not None:

            if not self._plotterPickerInitialized:
                self.plotter.enable_parallel_projection()
                self.plotter.enable_point_picking(show_message=False,
                                                  show_point=False,
                                                  pickable_window=True,
                                                  callback=lambda newPt: self._onSlicePointChanged())
                if isinstance(self.plotter, RemotePlotterProxy):
                    pass  # TODO: enable right button press response for remote plotter
                else:
                    set_mouse_event_for_picking(self.plotter, 'RightButtonPressEvent')
                self._plotterPickerInitialized = True

            if isinstance(self._activeSurf, str):
                surfKeys = [self._activeSurf]
            else:
                surfKeys = self._activeSurf

            if isinstance(self._surfColor, str):
                surfColors = [self._surfColor]
            else:
                surfColors = self._surfColor

            if not isinstance(self._surfOpacity, list):
                surfOpacities = [self._surfOpacity]
            else:
                surfOpacities = self._surfOpacity

            actors = []
            for iSurf, surfKey in enumerate(surfKeys):
                if getattr(self.session.headModel, surfKey) is not None:
                    if not self._surfPlotInitialized:
                        logger.debug('Initializing 3D plot')

                    actor = self._surfPlotter.add_mesh(mesh=getattr(self.session.headModel, surfKey),
                                                   color=surfColors[iSurf % len(surfColors)],
                                                   opacity=surfOpacities[iSurf % len(surfOpacities)],
                                                   name=self.label + '_' + surfKey + '_surf',
                                                   render=False,
                                                   reset_camera=False
                                                   )
                    actors.append(actor)

                    self._surfPlotInitialized = True

                    self.plotter.reset_camera()

            self._surfPlotActors.extend(actors)

        if True:
            logger.debug('Setting crosshairs for {} plot'.format(self.label))
            lineLength = 300  # TODO: scale by image size
            crosshairAxes = 'xyz'
            centerGapLength = 0
            for axis in crosshairAxes:
                mask = np.zeros((1, 3))
                mask[0, 'xyz'.index(axis)] = 1
                for iDir, dir in enumerate((-1, 1)):
                    pts = dir*np.asarray([centerGapLength/2, lineLength])[:, np.newaxis] * mask
                    if isinstance(self._normal, str):
                        pts += self._sliceOrigin
                    else:
                        viewToWorldTransf = composeTransform(self._normal, self._sliceOrigin)
                        pts = applyTransform(viewToWorldTransf, pts)
                    lineKey = 'Crosshair_{}_{}_{}'.format(self.label, axis, iDir)
                    if not self._plotterInitialized:
                        line = self._surfPlotter.add_lines(pts, color='#11DD11', width=2, name=lineKey)
                        with self._plotter.allowNonblockingCalls():
                            line.SetUseBounds(False)  # don't include for determining camera zoom, etc.
                        self._lineActors[lineKey] = line
                    else:
                        with self._plotter.allowNonblockingCalls():
                            logger.debug('Moving previous crosshairs')
                            line = self._lineActors[lineKey]
                            pts_pv = pv.lines_from_points(pts)
                            line.GetMapper().SetInputData(pts_pv)

        self._plotterInitialized = True

        with self._primaryPlotter.allowNonblockingCalls():
            self._primaryPlotter.render()

    def _onSlicePointChanged(self):
        pos = self.plotter.picked_point

        if True:
            # ray trace to find point closest to camera
            if self._pickableSurfs is None:
                if isinstance(self._activeSurf, str):
                    surfKeys = [self._activeSurf]
                else:
                    surfKeys = self._activeSurf
            else:
                surfKeys = self._pickableSurfs

            newPos = None

            for iSurf, surfKey in enumerate(surfKeys):
                mesh = getattr(self.session.headModel, surfKey)

                if False:
                    endPoint = pos
                else:
                    # selected pos may be "before" mesh, such that tracing to this endpoint would be insufficient to find mesh intersection
                    # so project even further
                    cameraToCrosshairsDist = np.linalg.norm(self.plotter.camera.position - self._sliceOrigin)
                    dir = np.asarray(pos) - np.asarray(self.plotter.camera.position)
                    dir /= np.linalg.norm(dir)
                    endPoint = pos + dir * cameraToCrosshairsDist * 2

                intersectionPoints, intersectionCells = mesh.ray_trace(
                    origin=self.plotter.camera.position,
                    end_point=endPoint,
                    first_point=True
                )
                if len(intersectionPoints) == 0:
                    continue  # no intersection
                if newPos is not None:
                    # already found an intersection, check if this one is closer
                    if np.linalg.norm(intersectionPoints - self.plotter.camera.position) < np.linalg.norm(newPos - self.plotter.camera.position):
                        newPos = intersectionPoints
                else:
                    newPos = intersectionPoints

            if newPos is None:
                # no intersections found, don't change slice origin
                return

            pos = newPos

        logger.debug('Slice point changed: {} {}'.format(self.label, pos))
        self.sliceOrigin = pos


import attrs
import logging
import typing as tp
import qtawesome as qta
from qtpy import QtGui

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K, logger
from NaviNIBS.Navigator.Model.DigitizedLocations import DigitizedLocations, DigitizedLocation
from NaviNIBS.util import makeStrUnique


logger = logging.getLogger()


@attrs.define(slots=False)
class DigitizedLocationsTableModel(CollectionTableModel[str, DigitizedLocations, DigitizedLocation]):
    _collection: DigitizedLocations = attrs.field(init=False, repr=False)

    _hasPlaceholderNewRow: bool = True
    _placeholderNewRowDefaults: dict[str, tp.Any] = attrs.field(factory=lambda: dict(key='<NewEntry>'))

    _checkIcon_planned: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.checkbox-marked-circle', color='blue'))
    _checkIcon_sampled: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.checkbox-marked-circle', color='green'))
    _xIcon: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.close-circle-outline', color='red'))

    def __attrs_post_init__(self):
        self._collection = self._session.digitizedLocations

        self._attrColumns = [
            'key',
            'type',
        ]

        self._derivedColumns = dict(
            plannedIsSet=self._getPlannedIsSet,
            sampledIsSet=self._getSampledIsSet
        )

        self._columns = self._attrColumns + list(self._derivedColumns.keys())

        self._decoratedColumns = [
            'plannedIsSet',
            'sampledIsSet'
        ]

        self._columnLabels = dict(
            plannedIsSet='Planned',
            sampledIsSet='Sampled'
        )

        self._editableColumns.extend(['key', 'type'])

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        self._addNewRowFromEditedPlaceholder = self.__addNewRowFromEditedPlaceholder

        super().__attrs_post_init__()

    def _getPlannedIsSet(self, key: tp.Optional[str]) -> tuple[tp.Optional[QtGui.QIcon], str]:
        if key is None:
            return None, ''
        if self._collection[key].plannedCoord is None:
            return self._xIcon, ''
        else:
            return self._checkIcon_planned, ''

    def _getSampledIsSet(self, key: tp.Optional[str]) -> tuple[tp.Optional[QtGui.QIcon], str]:
        if key is None:
            return None, ''
        if self._collection[key].sampledCoord is None:
            return self._xIcon, ''
        else:
            return self._checkIcon_sampled, ''

    def __addNewRowFromEditedPlaceholder(self, **kwargs) -> str | None:
        if 'key' not in kwargs:
            if len(kwargs) == 0:
                # nothing set, assume adding new row was cancelled
                return None
            kwargs['key'] = makeStrUnique('Loc', self._collection.keys())
        item = DigitizedLocation(**kwargs)
        if item.key in self._collection:
            # key already existed, reject
            return None
        self._collection.addItem(item)
        return item.key


import asyncio
import functools
import logging
import numpy as np
import pyvista as pv
from pyvista import _vtk
from qtpy import QtGui
import typing as tp
import weakref
if pv.__version__ <= '0.39.1':
    from pyvista.utilities.helpers import try_callback
else:
    from pyvista.core.utilities.misc import try_callback

from . import Actor

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, getActorUserTransform
from NaviNIBS.util import Transforms


logger = logging.getLogger(__name__)


def _launch_pick_event(interactor, event):
    """Create a Pick event based on coordinate or left-click."""
    click_x, click_y = interactor.GetEventPosition()
    click_z = 0

    picker = interactor.GetPicker()
    renderer = interactor.GetInteractorStyle()._parent()._plotter.renderer
    picker.Pick(click_x, click_y, click_z, renderer)


def set_mouse_event_for_picking(plotter: pv.BasePlotter, eventKey: str):
    """
    To be called after pyvista plotter.enable_*_picking(), allowing for
    picking to be triggered by alternate mouse events, such as left double click instead of single press
    :param eventKey: can be any relevant vtk event, e.g. 'LeftButtonPressEvent', 'RightButtonPressEvent'.
    :return: None

    Unfortunately, pyvista / Qt interactions mean that some events never reach VTK and so won't work here.
    For example, double click events don't work without some extra interactor modifications (see https://discourse.vtk.org/t/why-single-click-works-but-double-click-does-not/3599)
    """

    plotter.iren.interactor.AddObserver(
        eventKey,
        functools.partial(try_callback, _launch_pick_event),
    )


async def pickActor(plotter: pv.Plotter,
                    show: bool = False,
                    show_message: tp.Union[bool, str] = True,
                    style: str = 'wireframe',
                    line_width: float = 5,
                    color: str = 'pink',
                    font_size: int = 18,
                    left_clicking: bool = False,
                    **kwargs) -> tp.Optional[Actor]:
    """
    Adapted from pv.plotter.enable_mesh_picking but with a few improvements/modifications:
    - Supports highlighting selection even after UserTransform was set on original actor
    - Async call until pick is finished
    - Returns Actor instead of mesh PolyData
    """

    pickedActor = None
    event_pickFinished = asyncio.Event()

    selectActor = None

    def end_pick_call_back(picked, event):
        nonlocal pickedActor, selectActor

        is_valid_selection = False
        plotter_ = weakref.ref(plotter)

        pickedActor = picked.GetActor()
        if pickedActor:
            mesh = pickedActor.GetMapper().GetInput()
            is_valid_selection = True

        event_pickFinished.set()

        if not is_valid_selection:
            return

        selectActor = None

        if show:
            # Select the renderer where the mesh is added.
            active_renderer_index = plotter_().renderers._active_index
            for index in range(len(plotter.renderers)):
                renderer = plotter.renderers[index]
                for actor in renderer._actors.values():
                    try:
                        mapper = actor.GetMapper()
                    except Exception as e:
                        # not all non-mesh actors have mappers, just ignore
                        continue
                    if isinstance(mapper, _vtk.vtkDataSetMapper) and mapper.GetInput() == mesh:
                        loc = plotter_().renderers.index_to_loc(index)
                        plotter_().subplot(*loc)
                        break

            # Use try in case selection is empty or invalid
            try:
                selectActor = plotter_().add_mesh(
                    mesh,
                    name='_mesh_picking_selection',
                    style=style,
                    color=color,
                    line_width=line_width,
                    pickable=False,
                    reset_camera=False,
                    **kwargs,
                )
            except Exception as e:  # pragma: no cover
                logging.warning("Unable to show mesh when picking:\n\n%s", str(e))
            else:
                selectActor.SetUserTransform(pickedActor.GetUserTransform())

            # Reset to the active renderer.
            loc = plotter_().renderers.index_to_loc(active_renderer_index)
            plotter_().subplot(*loc)

            # render here prior to returning
            plotter_().render()

    if show_message:
        if show_message is True:
            show_message = "\nPress P to pick a single dataset under the mouse pointer."
            if left_clicking:
                show_message += "\nor click to select a dataset under the mouse pointer."

        textActor = plotter.add_text(str(show_message), font_size=font_size, name='_mesh_picking_message', color='black')
    else:
        textActor = None

    if left_clicking:
        leftClickObserver = plotter.iren.interactor.AddObserver(
            "LeftButtonPressEvent",
            functools.partial(try_callback, _launch_pick_event),
        )
    else:
        leftClickObserver = None

    logging.debug('About to wait')
    await asyncio.sleep(2.)  # TODO: debug, delete

    picker = _vtk.vtkPropPicker()
    endPickObserver = picker.AddObserver(_vtk.vtkCommand.EndPickEvent, end_pick_call_back)
    plotter.enable_trackball_style()
    plotter.iren.set_picker(picker)

    logging.debug('Waiting for pick to finish')
    await event_pickFinished.wait()
    logging.debug('Pick finished')

    # clean up
    if textActor is not None:
        plotter.remove_actor(textActor)

    if selectActor is not None:
        async def removeSelectActorAfterDelay():
            await asyncio.sleep(0.5)
            plotter.remove_actor(selectActor)
        asyncio.create_task(asyncTryAndLogExceptionOnError(removeSelectActorAfterDelay))

    picker.RemoveObserver(endPickObserver)
    if left_clicking:
        plotter.iren.interactor.RemoveObserver(leftClickObserver)

    return pickedActor


async def interactivelyMoveActor(plotter: pv.Plotter, actor: Actor, onNewTransf: tp.Callable[[_vtk.vtkTransform], None]):
    from vtkmodules.vtkInteractionWidgets import (
        vtkBoxWidget2,
    )

    wdgt = vtkBoxWidget2()
    wdgt.CreateDefaultRepresentation()
    wdgt.SetInteractor(plotter.iren.interactor)
    wdgt.SetCurrentRenderer(plotter.renderer)
    wdgt.SetScalingEnabled(False)
    wdgt.SetMoveFacesEnabled(False)
    handleRep = wdgt.GetRepresentation()
    palette = plotter.palette()
    handleRep.SetForegroundColor(*palette.color(QtGui.QPalette.Highlight).getRgbF()[:-1])
    handleRep.SetPlaceFactor(1.)
    if True:
        # temporarily reset transform to get bounds of actor in local space
        # TODO: do this with some other bounding box calculation rather than resetting transform
        actorTransf = actor.GetUserTransform()
        actor.SetUserTransform(_vtk.vtkTransform())
        localActorBounds = actor.GetBounds()
        actor.SetUserTransform(actorTransf)
        handleRep.PlaceWidget(localActorBounds)
        handleRep.SetTransform(actorTransf)
        plotter.render()
    else:
        handleRep.PlaceWidget(actor.GetBounds())
    #handleRep.SetOutlineFaceWires(True)
    handleRep.SetSnapToAxes(True)  # TODO: debug, delete
    wdgt.On()

    newTransfPending = asyncio.Event()
    newestTransf: tp.Optional[_vtk.vtkTransform] = None

    async def publishNewPosition():
        while True:
            await newTransfPending.wait()
            newTransfPending.clear()
            assert newestTransf is not None
            onNewTransf(newestTransf)
            await asyncio.sleep(0.1)  # don't publish updates more frequently than this

    publishTask = asyncio.create_task(asyncTryAndLogExceptionOnError(publishNewPosition))

    def interactionCallback(obj, event_type):
        nonlocal newestTransf
        transform = _vtk.vtkTransform()
        handleRep.GetTransform(transform)
        actor.SetUserTransform(transform)  # could be disabled if want to rely on caller to update the actor
        newestTransf = transform
        newTransfPending.set()

    # evtIDStr = _vtk.vtkCommand.GetStringFromEventId(_vtk.vtkCommand.GetEventIdFromString('InteractionEvent'))
    # logger.debug(f'evtIDStr: {evtIDStr}')
    wdgt.AddObserver('InteractionEvent', interactionCallback)

    if True:
        while True:
            await asyncio.sleep(1.)

    publishTask.cancel()
    # TODO: other clean-up





import typing as tp
from typing import ClassVar

import attrs
from NaviNIBS.util.pyvista.plotting import PrimaryLayeredPlotter, SecondaryLayeredPlotter
from NaviNIBS.util.pyvista.RemotePlotting import RemotePlotter as RP


class RemoteSecondaryLayeredPlotter(SecondaryLayeredPlotter, RP.RemotePlotterMixin):
    def __init__(self, *args, **kwargs):
        SecondaryLayeredPlotter.__init__(self, *args, **kwargs)
        RP.RemotePlotterMixin.__init__(self)


class RemotePrimaryLayeredPlotter(PrimaryLayeredPlotter, RP.RemotePlotterMixin):

    _createSecondaryPlotter: tp.Callable[..., RemoteSecondaryLayeredPlotter] = RemoteSecondaryLayeredPlotter
    """
    This allows creator to specify a different function to instantiate secondary plotters.
    First two arguments will be key and layer, and the rest will kwargs for the plotter.
    """

    def __init__(self, *args, **kwargs):

        self._createSecondaryPlotter = kwargs.pop('createSecondaryPlotter')  # required kwarg

        PrimaryLayeredPlotter.__init__(self, *args, **kwargs)
        RP.RemotePlotterMixin.__init__(self)

    def addLayeredPlotter(self, key: str, layer: tp.Optional[int] = None) -> int:
        """
        Note: this returns int layer rather than the plotter itself, unlike superclass
        """
        assert key not in self._secondaryPlotters
        self._secondaryPlotters[key] = self._createSecondaryPlotter(
            key, layer, mainPlotter=self, rendererLayer=layer)
        return self._secondaryPlotters[key].rendererLayer


@attrs.define
class RemoteSecondaryLayeredPlotManager(RP.RemotePlotManagerBase):
    _Plotter: ClassVar = RemoteSecondaryLayeredPlotter
    _plotter: SecondaryLayeredPlotter

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class RemotePrimaryLayeredPlotManager(RP.RemotePlotManager):

    _Plotter: ClassVar = RemotePrimaryLayeredPlotter
    _plotter: PrimaryLayeredPlotter | None = attrs.field(init=False, default=None)
    _secondaryPlotManagers: dict[str, RemoteSecondaryLayeredPlotManager] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        self._plotterKwargs['createSecondaryPlotter'] = self._createSecondaryPlot

        super().__attrs_post_init__()

    def _createSecondaryPlot(self, key: str, layer: int | None, **kwargs) -> SecondaryLayeredPlotter:
        self._secondaryPlotManagers[key] = RemoteSecondaryLayeredPlotManager(
            plotter=RemoteSecondaryLayeredPlotter(**kwargs),
            parentLayout=self._parentLayout,
        )
        return self._secondaryPlotManagers[key].plotter

    async def _handleMsg(self, msg):
        # handle first argument that represents primary/secondary plotter key
        plotterKey = msg[0]

        if plotterKey == '__primary':
            return await super()._handleMsg(msg[1:])
        else:
            return await self._secondaryPlotManagers[plotterKey]._handleMsg(msg[1:])


@attrs.define
class RemoteLayeredPlotterApp(RP.RemotePlotterApp):
    _plotManager: RemotePrimaryLayeredPlotManager = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _initPlotManager(self):
        assert self._plotManager is None
        self._plotManager = RemotePrimaryLayeredPlotManager(
            reqPort=self._reqPort,
            repPort=self._repPort,
            parentLayout=self._rootWdgt.layout(),
            plotterKwargs=self._plotterKwargs,
        )


import json
import jsbeautifier


def beautifyJSON(o: str) -> str:
    """
    Python JSON module doesn't make it simple to do prettier formatting with most entries being indented but keeping
     things like simple, short arrays on single lines. So use jsbeautifier to process after dumping.
    See https://stackoverflow.com/questions/21866774/pretty-print-json-dumps
    Note that this is not particularly efficient or lightweight of a dependency, but perhaps better than hacky
     subclassing of JSONEncoder as in http://stackoverflow.com/a/17684652
    """
    opts = jsbeautifier.default_options()
    opts.indent_size = 2
    return jsbeautifier.beautify(o, opts)


def jsonPrettyDumps(o) -> str:
    return beautifyJSON(json.dumps(o))



import typing as tp
from qtpy import QtWidgets, QtGui, QtCore


class IconWidget(QtWidgets.QLabel):
    """
    Adapted from https://stackoverflow.com/a/64279374
    """
    _icon: QtGui.QIcon
    _size: QtCore.QSize

    def __init__(self, icon: QtGui.QIcon, size: tp.Optional[QtCore.QSize] = None):
        if size is None:
            size = QtCore.QSize(16, 16)

        self._size = size
        self._icon = icon

        super().__init__()

        self.setPixmap(icon.pixmap(size))

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, newIcon: QtGui.QIcon):
        self._icon = newIcon
        self.setPixmap(self._icon.pixmap(self._size))


import asyncio
import attrs
import inspect
import logging
import numpy as np
import pyvista as pv
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Navigator.GUI.EditWindows.ToolCalibrationWindow import ToolCalibrationWindow
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Transforms import invertTransform, concatenateTransforms
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments
from NaviNIBS.util.pyvista import DefaultPrimaryLayeredPlotter, DefaultSecondaryLayeredPlotter, RemotePlotterProxy


logger = logging.getLogger(__name__)


@attrs.define
class CoilCalibrationWithPlateWindow(ToolCalibrationWindow):
    """
    Calibrate a coil by aligning to a calibration plate
    """

    _pendingNewTransf: tp.Optional[np.ndarray] = attrs.field(init=False, default=None)  # result of calibration, not yet saved to tool

    _instructions: QtWidgets.QTextEdit = attrs.field(init=False)

    _plotter: DefaultPrimaryLayeredPlotter = attrs.field(init=False)
    _plotterUpperLayer: DefaultSecondaryLayeredPlotter = attrs.field(init=False)
    _plotterLowerLayer: DefaultSecondaryLayeredPlotter = attrs.field(init=False)

    _coilToolActor: Actor | None = attrs.field(init=False, default=None)
    _coilTrackerActor: Actor | None = attrs.field(init=False, default=None)
    _calibrationPlateToolActor: Actor | None = attrs.field(init=False, default=None)
    _calibrationPlateTrackerActor: Actor | None = attrs.field(init=False, default=None)
    _coilAxesActor: Actor | None = attrs.field(init=False, default=None)
    _calibrationPlateAxesActor: Actor | None = attrs.field(init=False, default=None)

    _calibrateBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _revertBtn: QtWidgets.QPushButton = attrs.field(init=False)

    _finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    # TODO: add error metrics

    # TODO: add checkboxes to optionally hide some actors

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        splitter = QtWidgets.QSplitter()
        self._wdgt.layout().addWidget(splitter)

        self._instructions = QtWidgets.QTextEdit('')
        self._instructions.setReadOnly(True)
        self._instructions.setMarkdown(inspect.cleandoc("""
                    # Coil calibration with calibration plate
                    ## Instructions
                    1. Hold calibration plate in location visible to camera.
                    2. Position coil on plate, with top face of plate touching bottom face of coil.
                    3. Align center of coil with center of plate, and align coil axes with plate axes.
                    4. Press "Calibrate coil".
                    5. Check calibration by removing coil, then repositioning in same place and evaluating visual alignment and error metrics.
                    6. Repeat calibration if necessary.
                    7. Close this window to accept the calibration. 
                """))
        splitter.addWidget(self._instructions)

        splitRight = QtWidgets.QWidget()
        splitRight.setLayout(QtWidgets.QVBoxLayout())
        splitter.addWidget(splitRight)
        splitter.setStretchFactor(1, 2)

        self._plotter = DefaultPrimaryLayeredPlotter()
        splitRight.layout().addWidget(self._plotter)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QHBoxLayout())
        splitRight.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Calibrate coil')
        btn.clicked.connect(lambda checked: self.calibrate())
        btn.setEnabled(False)
        btnContainer.layout().addWidget(btn)
        self._calibrateBtn = btn

        btn = QtWidgets.QPushButton('Revert to original calibration')
        btn.clicked.connect(lambda checked: self.revert())
        btn.setEnabled(False)
        btnContainer.layout().addWidget(btn)
        self._revertBtn = btn

        # TODO: add error metrics (distance from previous calibration) for checking if need to recalibrate

        self.sigFinished.connect(self._onDialogFinished)

        self._wdgt.resize(QtCore.QSize(900, 700))

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):

        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        self._plotterUpperLayer = self._plotter.addLayeredPlotter(key='AxesIndicators', layer=1)
        self._plotterLowerLayer = self._plotter.addLayeredPlotter(key='Meshes', layer=0)

        self._finishedAsyncInit.set()

        self._onLatestPositionsChanged()

    def calibrate(self):
        # TODO: spin-wait positions client to make sure we have most up-to-date information
        coilTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey)
        calibrationPlateTrackerToCameraTransf = self._positionsClient.getLatestTransf(self._session.tools.calibrationPlate.trackerKey)
        calibrationPlateToTrackerTransf = self._session.tools.calibrationPlate.toolToTrackerTransf

        self._pendingNewTransf = concatenateTransforms([
            calibrationPlateToTrackerTransf,
            calibrationPlateTrackerToCameraTransf,
            invertTransform(coilTrackerToCameraTransf)
        ])

        logger.info('Calibrated {} transform: {}'.format(self._toolKeyToCalibrate, self._pendingNewTransf))

        self._revertBtn.setEnabled(True)

        self._onLatestPositionsChanged()

    def revert(self):
        self._pendingNewTransf = None
        self._revertBtn.setEnabled(False)
        self._onLatestPositionsChanged()

    def _onDialogFinished(self, wasAccepted: bool):
        if self._pendingNewTransf is not None:
            self._session.tools[self._toolKeyToCalibrate].toolToTrackerTransf = self._pendingNewTransf
            logger.info('Saved {} calibration: {}'.format(self._toolKeyToCalibrate, self.toolToCalibrate.toolToTrackerTransf))

        self._plotter.close()

    def _resetCamera(self):
        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not yet ready
            return

        yOffset = -35  # since coil will typically extend to bottom of screen after calibration,
        # offset view toward top to reduce excessive whitespace
        self._plotter.camera.focal_point = (0, yOffset, 0)
        self._plotter.camera.position = (0, yOffset, 500)
        self._plotter.camera.up = (0, 1, 0)
        self._plotter.camera.view_angle = 30.
        self._plotter.camera.clipping_range = (1, 1400)
        self._plotter.render()

    def _onLatestPositionsChanged(self):

        if not self._finishedAsyncInit.is_set():
            # plotter not yet ready
            return

        needsRender = False
        doResetCamera = False

        super()._onLatestPositionsChanged()

        calibrationPlate = self._session.tools.calibrationPlate
        if self.toolToCalibrate.isActive and calibrationPlate is not None \
                and self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey, None) is not None \
                and self._positionsClient.getLatestTransf(calibrationPlate.trackerKey, None) is not None:
            self._calibrateBtn.setEnabled(True)
        else:
            self._calibrateBtn.setEnabled(False)

        calibPlate = self._session.tools.calibrationPlate
        if calibPlate is None:
            calibPlateTrackerToCameraTransf = None
        else:
            calibPlateTrackerToCameraTransf = self._positionsClient.getLatestTransf(calibPlate.trackerKey, None)

        if calibPlateTrackerToCameraTransf is None:
            # without calibration plate pose, can't show anything in view that is relative to calibration plate
            for actor in (
                self._calibrationPlateAxesActor,
                self._calibrationPlateToolActor,
                self._calibrationPlateTrackerActor,
                self._coilAxesActor,
                self._coilToolActor,
                self._coilTrackerActor,
            ):
                if actor is not None and actor.GetVisibility():
                    actor.VisibilityOff()
                    needsRender = True

            if needsRender:
                self._plotter.render()
            return

        else:
            for actor in (
                    self._calibrationPlateAxesActor,
                    self._calibrationPlateToolActor,
                    self._calibrationPlateTrackerActor
            ):
                if actor is not None and not actor.GetVisibility():
                    actor.VisibilityOn()
                    needsRender = True

        if self._calibrationPlateAxesActor is None:
            self._calibrationPlateAxesActor = self._plotterUpperLayer.add_axes_marker(
                xlabel='Ref X\n',
                ylabel='Ref Y\n',
                zlabel='',
                label_size=(0.1, 0.1),
                line_width=3,
                total_length=(50, 50, 50)
            )
            # no transform needed here
            needsRender = True
            doResetCamera = True

        if self._calibrationPlateToolActor is None:
            if calibPlate.toolSurf is not None:
                mesh = calibPlate.toolSurf
                meshColor = calibPlate.toolColor

                self._calibrationPlateToolActor = self._plotterLowerLayer.addMesh(
                    mesh=mesh,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.5
                )

                setActorUserTransform(self._calibrationPlateToolActor, calibPlate.toolStlToToolTransf)
                needsRender = True
                doResetCamera = True

        if self._calibrationPlateTrackerActor is None:
            if calibPlate.trackerSurf is not None:
                mesh = calibPlate.trackerSurf
                meshColor = calibPlate.trackerColor
                self._calibrationPlateTrackerActor = self._plotterLowerLayer.addMesh(
                    mesh=mesh,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.5
                )
                setActorUserTransform(self._calibrationPlateTrackerActor, concatenateTransforms([
                    calibPlate.trackerStlToTrackerTransf,
                    invertTransform(calibPlate.toolToTrackerTransf),
                ]))
                needsRender = True
                doResetCamera = True

        coil = self.toolToCalibrate
        if coil is None:
            coilTrackerToCameraTransf = None
        else:
            coilTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey, None)

        if coilTrackerToCameraTransf is None:
            # without coil pose, can't show any of the coil-related actors
            for actor in (
                    self._coilAxesActor,
                    self._coilToolActor,
                    self._coilTrackerActor,
            ):
                if actor is not None and actor.GetVisibility():
                    actor.VisibilityOff()
                    needsRender = True

            if doResetCamera:
                self._resetCamera()

            if needsRender:
                self._plotter.render()
            return

        else:
            for actor in (
                    self._coilAxesActor,
                    self._coilToolActor,
                    self._coilTrackerActor,
            ):
                if actor is not None and not actor.GetVisibility():
                    actor.VisibilityOn()

        if self._coilAxesActor is None:
            self._coilAxesActor = self._plotterUpperLayer.add_axes_marker(
                xlabel='\nCoil X',
                ylabel='\nCoil Y',
                zlabel='',
                label_size=(0.1, 0.1),
                total_length=(50, 50, 50)
            )
            doResetCamera = True

        if self._coilToolActor is None:
            if coil.toolSurf is not None:
                mesh = coil.toolSurf
                meshColor = coil.toolColor
                self._coilToolActor = self._plotterLowerLayer.addMesh(
                    mesh=mesh,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.5
                )
                doResetCamera = True

        if self._coilTrackerActor is None:
            if coil.trackerSurf is not None:
                mesh = coil.trackerSurf
                meshColor = coil.trackerColor
                scalars = None
                if meshColor is None:
                    if len(mesh.array_names) > 0:
                        meshColor = None  # use color from surf file
                        scalars = mesh.array_names[-1]
                    else:
                        meshColor = '#444444'  # default color if nothing else provided
                self._coilTrackerActor = self._plotterLowerLayer.addMesh(
                    mesh=mesh,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.5)
                doResetCamera = True

        coilTrackerToCalibToolTransf = concatenateTransforms([
            coilTrackerToCameraTransf,
            invertTransform(calibPlateTrackerToCameraTransf),
            invertTransform(calibPlate.toolToTrackerTransf)
        ])
        if self._coilTrackerActor is not None:
            setActorUserTransform(self._coilTrackerActor, concatenateTransforms([
                coil.trackerStlToTrackerTransf,
                coilTrackerToCalibToolTransf
            ]))

        coilToolToCalibToolTransf = coilTrackerToCalibToolTransf.copy()
        if self._pendingNewTransf is None:
            if coil.toolToTrackerTransf is None:
                # assume identity transform initially
                pass
            else:
                coilToolToCalibToolTransf = concatenateTransforms([
                    coil.toolToTrackerTransf,
                    coilTrackerToCalibToolTransf
                ])
        else:
            coilToolToCalibToolTransf = concatenateTransforms([
                self._pendingNewTransf,
                coilTrackerToCalibToolTransf
            ])
        setActorUserTransform(self._coilAxesActor, coilToolToCalibToolTransf)
        if self._coilToolActor is not None:
            setActorUserTransform(self._coilToolActor, concatenateTransforms([
                coil.toolStlToToolTransf,
                coilToolToCalibToolTransf
            ]))

        if doResetCamera:
            self._resetCamera()

        self._plotter.render()


@attrs.define
class CoilCalibrationWithPointerWindow(ToolCalibrationWindow):
    """
    Calibrate a coil by pointing to known locations on the coil with pointer
    """
    def __attrs_post_init__(self):
        raise NotImplementedError  # TODO


import asyncio
import attrs
import json
import logging
import numpy as np
import pytransform3d.rotations as ptr
from qtpy import QtWidgets, QtGui, QtCore
from skspatial.objects import Vector
import typing as tp

from NaviNIBS.Navigator.Model.Targets import Target
from NaviNIBS.Navigator.GUI.CollectionModels.TargetsTableModel import TargetsTableModel, FullTargetsTableModel
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.Calculations import getClosestPointToPointOnMesh, calculateCoilToMRITransfFromTargetEntryAngle
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import applyTransform, invertTransform, composeTransform, concatenateTransforms, applyDirectionTransform, calculateRotationMatrixFromVectorToVector
from NaviNIBS.util.GUI.QDial import AngleDial
from NaviNIBS.util.GUI.QScrollContainer import QScrollContainer
from NaviNIBS.util.GUI.QMouseWheelAdjustmentGuard import preventAnnoyingScrollBehaviour

logger = logging.getLogger(__name__)


@attrs.define(kw_only=True)
class EditGridWidget:
    _session: Session = attrs.field(repr=False)

    _wdgt: QtWidgets.QWidget = attrs.field(factory=lambda: QtWidgets.QGroupBox('Edit grid'))
    _scroll: QScrollContainer = attrs.field(init=False)

    _targetComboBox: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)
    _targetsModel: FullTargetsTableModel = attrs.field(init=False)
    _preChangeTargetComboBoxIndex: list[int] = attrs.field(init=False, factory=list)

    _gridPrimaryAngleWdgt: AngleDial = attrs.field(init=False)

    _gridSpacingAtDepthWdgt: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)
    _gridPivotDepth: QtWidgets.QDoubleSpinBox = attrs.field(init=False, factory=QtWidgets.QDoubleSpinBox)
    _gridDepthMethodWdgt: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)
    _gridEntryAngleMethodWdgt: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)

    _seedTarget: Target | None = attrs.field(init=False, default=None)

    _disableWidgetsWhenNoTarget: list[QtWidgets.QWidget] = attrs.field(init=False, factory=list)

    _gridWidthWdgts: tuple[QtWidgets.QDoubleSpinBox, QtWidgets.QDoubleSpinBox] = attrs.field(init=False)
    _gridNWdgts: tuple[QtWidgets.QSpinBox, QtWidgets.QSpinBox] = attrs.field(init=False)

    _gridHandleAngleWdgts: tuple[AngleDial, AngleDial] = attrs.field(init=False)
    _gridHandleAngleNWdgt: QtWidgets.QSpinBox = attrs.field(init=False)

    _pendingGridTargetKeys: list[str] = attrs.field(init=False, factory=list)

    _gridNeedsUpdate: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        outerLayout = QtWidgets.QVBoxLayout()
        self._wdgt.setLayout(outerLayout)
        outerLayout.setContentsMargins(0, 0, 0, 0)

        layout = QtWidgets.QFormLayout()
        self._scroll = QScrollContainer(innerContainerLayout=layout)
        self._scroll.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        outerLayout.addWidget(self._scroll.scrollArea)

        self._scroll.scrollArea.setSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                              QtWidgets.QSizePolicy.Preferred)

        self._targetsModel = FullTargetsTableModel(session=self._session)
        self._targetComboBox.setModel(self._targetsModel)
        if True:
            # because setting targets from empty to non-empty also sets current index to 0, need to monitor collection directly to block this
            self._targetsModel.collection.sigItemsAboutToChange.connect(
                self._onTargetsCollectionAboutToChange, priority=1)
            self._targetsModel.collection.sigItemsChanged.connect(
                self._onTargetsCollectionChanged, priority=-1)
        self._targetComboBox.setCurrentIndex(-1)
        self._targetComboBox.currentIndexChanged.connect(self._onTargetComboBoxCurrentIndexChanged)
        layout.addRow('Seed target:', self._targetComboBox)

        # this "primary angle" defines angle of grid X axis relative to seed target's coil X axis
        self._gridPrimaryAngleWdgt = AngleDial(
            centerAngle=0,
            offsetAngle=-90,
            doInvert=True
        )
        self._gridPrimaryAngleWdgt.sigValueChanged.connect(self._onGridPrimaryAngleChanged)
        layout.addRow('Grid angle offset:', self._gridPrimaryAngleWdgt.wdgt)
        self._disableWidgetsWhenNoTarget.append(self._gridPrimaryAngleWdgt.wdgt)

        self._gridSpacingAtDepthWdgt.addItems(['Coil', 'Entry', 'Target'])
        self._gridSpacingAtDepthWdgt.setCurrentIndex(2)
        self._gridSpacingAtDepthWdgt.currentIndexChanged.connect(self._onGridSpacingAtDepthChanged)
        layout.addRow('Grid at depth of:', self._gridSpacingAtDepthWdgt)
        self._disableWidgetsWhenNoTarget.append(self._gridSpacingAtDepthWdgt)

        self._gridPivotDepth.setRange(1, 1000)
        self._gridPivotDepth.setSingleStep(1)
        self._gridPivotDepth.setDecimals(1)
        self._gridPivotDepth.setSuffix(' mm')
        self._gridPivotDepth.setValue(60.)
        self._gridPivotDepth.valueChanged.connect(self._onGridPivotDepthChanged)
        # self._gridPivotDepth.setKeyboardTracking(False)
        layout.addRow('Grid pivot depth:', self._gridPivotDepth)
        self._disableWidgetsWhenNoTarget.append(self._gridPivotDepth)

        self._gridDepthMethodWdgt.addItems(['from pivot', 'from skin'])
        self._gridDepthMethodWdgt.setCurrentIndex(1)
        self._gridDepthMethodWdgt.currentIndexChanged.connect(self._onGridDepthMethodChanged)
        layout.addRow('Grid depth adjustment:', self._gridDepthMethodWdgt)
        self._disableWidgetsWhenNoTarget.append(self._gridDepthMethodWdgt)

        self._gridWidthWdgts = (
            QtWidgets.QDoubleSpinBox(),
            QtWidgets.QDoubleSpinBox()
        )

        self._gridNWdgts = (
            QtWidgets.QSpinBox(),
            QtWidgets.QSpinBox()
        )

        for iXY, (gridWidthWdgt, gridNWdgt) in enumerate(zip(self._gridWidthWdgts, self._gridNWdgts)):
            preventAnnoyingScrollBehaviour(gridWidthWdgt)
            # gridWidthWdgt.setKeyboardTracking(False)
            gridWidthWdgt.setRange(0, 1000)
            gridWidthWdgt.setSingleStep(1)
            gridWidthWdgt.setDecimals(1)
            gridWidthWdgt.setSuffix(' mm')
            gridWidthWdgt.setValue(20.)
            gridWidthWdgt.valueChanged.connect(self._onGridWidthChanged)
            layout.addRow(f'Grid width {"XY"[iXY]}:', gridWidthWdgt)
            self._disableWidgetsWhenNoTarget.append(gridWidthWdgt)

            preventAnnoyingScrollBehaviour(gridNWdgt)
            # gridNWdgt.setKeyboardTracking(False)
            gridNWdgt.setRange(1, 1000)
            gridNWdgt.setSingleStep(1)
            gridNWdgt.valueChanged.connect(self._onGridNChanged)
            layout.addRow(f'Grid N {"XY"[iXY]}:', gridNWdgt)
            self._disableWidgetsWhenNoTarget.append(gridNWdgt)

        # noinspection PyTypeChecker
        self._gridHandleAngleWdgts = tuple(
            AngleDial(centerAngle=0,
                      offsetAngle=-90,
                      doInvert=True) for _ in range(2))
        for iAngle, (angleWdgt, angleLabel) in enumerate(zip(self._gridHandleAngleWdgts, ('start', 'end'))):
            angleWdgt.sigValueChanged.connect(self._onGridHandleAngleSpanChanged)
            layout.addRow(f'Handle angle {angleLabel}', angleWdgt.wdgt)
            self._disableWidgetsWhenNoTarget.append(angleWdgt.wdgt)

        self._gridHandleAngleNWdgt = QtWidgets.QSpinBox()
        self._gridHandleAngleNWdgt.setRange(1, 1000)
        self._gridHandleAngleNWdgt.setSingleStep(1)
        self._gridHandleAngleNWdgt.valueChanged.connect(self._onGridNChanged)
        layout.addRow(f'Grid handle angle N', self._gridHandleAngleNWdgt)
        self._disableWidgetsWhenNoTarget.append(self._gridHandleAngleNWdgt)

        btnContainer = QtWidgets.QWidget()
        btnLayout = QtWidgets.QHBoxLayout()
        btnContainer.setLayout(btnLayout)
        btnLayout.setContentsMargins(0, 0, 0, 0)

        cancelBtn = QtWidgets.QPushButton('Cancel')
        cancelBtn.clicked.connect(self._onCancelBtnClicked)
        btnLayout.addWidget(cancelBtn)
        finishBtn = QtWidgets.QPushButton('Finish')
        finishBtn.clicked.connect(self._onFinishBtnClicked)
        btnLayout.addWidget(finishBtn)
        layout.addWidget(btnContainer)
        self._disableWidgetsWhenNoTarget.extend([cancelBtn, finishBtn])

        for wdgt in self._disableWidgetsWhenNoTarget:
            wdgt.setEnabled(self._seedTarget is not None)

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_updateGrid))

        self._gridNeedsUpdate.set()

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def seedTarget(self):
        return self._seedTarget

    @seedTarget.setter
    def seedTarget(self, newTarget: Target | None):
        if self._seedTarget is newTarget:
            return

        if self._seedTarget is not None:
            self._seedTarget.sigItemChanged.disconnect(self._onSeedTargetItemChanged)

        self._deleteAnyPendingGridTargets()

        self._seedTarget = newTarget

        if self._seedTarget is not None:
            index = self._targetsModel.getIndexFromCollectionItemKey(newTarget.key)
            if index is not None:
                self._targetComboBox.setCurrentIndex(index)

            self._seedTarget.sigItemChanged.connect(self._onSeedTargetItemChanged)

        for wdgt in self._disableWidgetsWhenNoTarget:
            wdgt.setEnabled(self._seedTarget is not None)

        if self._seedTarget is not None:
            self._gridNeedsUpdate.set()

    async def _loop_updateGrid(self):
        while True:
            await self._gridNeedsUpdate.wait()
            await asyncio.sleep(0.1)  # rate-limit
            self._gridNeedsUpdate.clear()
            self._regenerateGrid()

    def _onTargetsCollectionAboutToChange(self, *args, **kwargs):
        if len(self._preChangeTargetComboBoxIndex) == 0:
            # don't respond to combo box index changes during a targets change,
            # since if the starting selection is empty it will force reset to non-empty
            self._targetComboBox.currentIndexChanged.disconnect(
                self._onTargetComboBoxCurrentIndexChanged)
        self._preChangeTargetComboBoxIndex.append(self._targetComboBox.currentIndex())

    def _onTargetsCollectionChanged(self, *args, **kwargs):
        if len(self._preChangeTargetComboBoxIndex) > 0:
            # one a change is complete, restore previous index
            # if it was empty, and respond to any other change
            prevIndex = self._preChangeTargetComboBoxIndex.pop()

            if len(self._preChangeTargetComboBoxIndex) == 0:
                if prevIndex == -1:
                    self._targetComboBox.setCurrentIndex(-1)
                elif prevIndex != self._targetComboBox.currentIndex():
                    self._onTargetComboBoxCurrentIndexChanged(self._targetComboBox.currentIndex())

                self._targetComboBox.currentIndexChanged.connect(
                    self._onTargetComboBoxCurrentIndexChanged)

    def _deleteAnyPendingGridTargets(self):
        if len(self._pendingGridTargetKeys) > 0:
            logger.debug(f'Deleting pending grid targets: {self._pendingGridTargetKeys}')
            self._session.targets.deleteItems(self._pendingGridTargetKeys)
            self._pendingGridTargetKeys.clear()

    def _regenerateGrid(self):
        logger.debug('Regenerating grid')
        # TODO: instead of deleting everything and recreating, edit / repurpose existing targets
        self._deleteAnyPendingGridTargets()

        #return  # TODO: debug, delete

        if self._seedTarget is None:
            return

        logger.info(f'Generating grid for seedTarget {self._seedTarget.key}')

        gridWidthX, gridWidthY = (wdgt.value() for wdgt in self._gridWidthWdgts)

        gridNX, gridNY = (wdgt.value() for wdgt in self._gridNWdgts)

        gridHandleAngleStart, gridHandleAngleStop = (wdgt.value for wdgt in self._gridHandleAngleWdgts)

        gridNAngle = self._gridHandleAngleNWdgt.value()

        depthMode = self._gridSpacingAtDepthWdgt.currentText()
        match depthMode:
            case 'Coil':
                refOrigin = self._seedTarget.entryCoordPlusDepthOffset
                refDepthFromSeedCoil = 0.
                refDepthFromSeedEntry = self._seedTarget.depthOffset
                refDepthFromSeedTarget = np.linalg.norm(self._seedTarget.entryCoord - self._seedTarget.targetCoord) + self._seedTarget.depthOffset
            case 'Entry':
                refOrigin = self._seedTarget.entryCoord
                refDepthFromSeedCoil = -self._seedTarget.depthOffset
                refDepthFromSeedEntry = 0.
                refDepthFromSeedTarget = np.linalg.norm(self._seedTarget.entryCoord - self._seedTarget.targetCoord)
            case 'Target':
                refOrigin = self._seedTarget.targetCoord
                refDepthFromSeedCoil = -np.linalg.norm(self._seedTarget.entryCoord - self._seedTarget.targetCoord) - self.seedTarget.depthOffset
                refDepthFromSeedEntry = -np.linalg.norm(self._seedTarget.entryCoord - self._seedTarget.targetCoord)
                refDepthFromSeedTarget = 0.
            case _:
                raise NotImplementedError

        seedCoilToMRITransf = self._seedTarget.coilToMRITransf
        seedCoilToUnrotSeedCoil = composeTransform(ptr.active_matrix_from_angle(2, -np.deg2rad(self._gridPrimaryAngleWdgt.value)))  # TODO: check sign of angle
        gridSpaceToSeedCoilTransf = composeTransform(ptr.active_matrix_from_angle(2, np.deg2rad(self._gridPrimaryAngleWdgt.value)),  # TODO: check sign of angle
                                                     np.asarray([0, 0, refDepthFromSeedCoil]))
        gridSpaceToMRITransf = concatenateTransforms([gridSpaceToSeedCoilTransf, seedCoilToMRITransf])

        entryDir = Vector(self._seedTarget.entryCoord - self._seedTarget.targetCoord).unit()

        pivotDepth = self._gridPivotDepth.value()  # in mm, dist from seed at grid depth to "common" origin

        # extraTransf = np.eye(4)
        # extraTransf[:3, 3] = refDepthFromSeedCoil
        # pivotOrigin_MRISpace = applyTransform(
        #     concatenateTransforms((extraTransf, self._seedTarget.coilToMRITransf)),
        #     np.asarray([0, 0, -pivotDepth]))

        # TODO: update grid spacing below to be defined based on arc lengths around pivot origin, instead of in a 2D plane

        totalThetaX = gridWidthX / pivotDepth
        assert totalThetaX < 2*np.pi, 'Grid width too large for given pivot depth'

        totalThetaY = gridWidthY / pivotDepth
        assert totalThetaY < 2*np.pi, 'Grid width too large for given pivot depth'

        if gridNX == 1:
            thetaXs = [0.]
        else:
            thetaXs = np.linspace(-totalThetaX / 2, totalThetaX / 2, gridNX)
        if gridNY == 1:
            thetaYs = [0.]
        else:
            thetaYs = np.linspace(-totalThetaY / 2, totalThetaY / 2, gridNY)
        if gridNAngle == 1:
            aCoords_seed = [np.mean([gridHandleAngleStart, gridHandleAngleStop])]
        else:
            aCoords_seed = np.linspace(gridHandleAngleStart, gridHandleAngleStop, gridNAngle)

        numPoints = gridNX * gridNY * gridNAngle
        newToGridSpaceTransfs = np.full((numPoints, 4, 4), np.nan)
        gridHandleAngles = np.full((numPoints,), np.nan)

        for iX, thetaX in enumerate(thetaXs):
            for iY, thetaY in enumerate(thetaYs):
                transf_gridSpaceToPivot = np.eye(4)
                transf_gridSpaceToPivot[2, 3] = pivotDepth

                rot = ptr.active_matrix_from_extrinsic_euler_yxy((-thetaX, -thetaY, 0))  # TODO: check sign
                transf_pivotToPivoted = composeTransform(rot)

                transf_pivotedToNewUnrot = np.eye(4)
                transf_pivotedToNewUnrot[2, 3] = -pivotDepth

                for iA, aCoord in enumerate(aCoords_seed):
                    transf_newUnrotToNew = concatenateTransforms([
                        invertTransform(seedCoilToUnrotSeedCoil),
                        composeTransform(ptr.active_matrix_from_angle(2, -np.deg2rad(aCoord))),  # TODO: check sign of angle
                    ])

                    newToGridSpaceTransfs[iX*gridNY*gridNAngle + iY*gridNAngle + iA] = invertTransform(concatenateTransforms(
                        [transf_gridSpaceToPivot, transf_pivotToPivoted, transf_pivotedToNewUnrot, transf_newUnrotToNew]))

                    gridHandleAngles[iX*gridNY*gridNAngle + iY*gridNAngle + iA] = aCoord

        newToMRISpaceTransfs = np.full((numPoints, 4, 4), np.nan)
        for i in range(numPoints):
            newToMRISpaceTransfs[i] = concatenateTransforms([newToGridSpaceTransfs[i], gridSpaceToMRITransf])

        newCoilToNewTransf = np.eye(4)
        newCoilToNewTransf[2, 3] = -refDepthFromSeedCoil  # before any additional depth correction (e.g. before matching to new skin depth)

        newEntryToNewTransf = np.eye(4)
        newEntryToNewTransf[2, 3] = -refDepthFromSeedEntry

        newTargetToNewTransf = np.eye(4)
        newTargetToNewTransf[2, 3] = -refDepthFromSeedTarget

        for i in range(numPoints):
            uniqueTargetKey = makeStrUnique(baseStr=f'{self._seedTarget.key} grid point {i+1}', # TODO: include X and Y indices separately in grid key
                                            existingStrs=self._session.targets.keys(),
                                            delimiter='#')

            newCoilToMRITransf = concatenateTransforms([newCoilToNewTransf, newToMRISpaceTransfs[i]])
            entryCoord_MRISpace = applyTransform((newEntryToNewTransf, newToMRISpaceTransfs[i]), np.asarray([0, 0, 0]))
            targetCoord_MRISpace = applyTransform((newTargetToNewTransf, newToMRISpaceTransfs[i]), np.asarray([0, 0, 0]))

            newTarget = Target(
                session=self._session,
                coilToMRITransf=newCoilToMRITransf,
                targetCoord=targetCoord_MRISpace,
                entryCoord=entryCoord_MRISpace,
                depthOffset=self._seedTarget.depthOffset,
                key=uniqueTargetKey,
                angle=self._seedTarget.angle + gridHandleAngles[i],  # TODO: check sign of offset, and note that this is approximate due to pivot angles
                color=self._seedTarget.color,
            )
            logger.debug(f'New target: {newTarget}')
            self._pendingGridTargetKeys.append(newTarget.key)
            self._session.targets.addItem(newTarget)

    def _onSeedTargetItemChanged(self, item: Target, attribsChanged: list[str] | None = None):

        if attribsChanged is not None and all(attrib in
                                              ('isVisible',
                                               'isSelected',
                                               'isHistorical',
                                               'mayBeADependency',
                                               'isSelected')
                                              for attrib in attribsChanged):
            # can ignore these changes
            return

        self._gridNeedsUpdate.set()

    def _onTargetComboBoxCurrentIndexChanged(self, index: int):
        if index == -1:
            self.seedTarget = None
            return

        seedTargetKey = self._targetsModel.getCollectionItemKeyFromIndex(index)
        self.seedTarget = self._session.targets[seedTargetKey]

    def _onGridPrimaryAngleChanged(self, angle: float):
        self._gridNeedsUpdate.set()

    def _onGridSpacingAtDepthChanged(self, index: int):
        self._gridNeedsUpdate.set()

    def _onGridPivotDepthChanged(self, value: float):
        self._gridNeedsUpdate.set()

    def _onGridDepthMethodChanged(self, index: int):
        self._gridNeedsUpdate.set()

    def _onGridEntryAngleMethodChanged(self, index: int):
        self._gridNeedsUpdate.set()

    def _onGridWidthChanged(self, value: float):
        self._gridNeedsUpdate.set()

    def _onGridNChanged(self, value: int):
        self._gridNeedsUpdate.set()

    def _onGridHandleAngleSpanChanged(self, *args):
        self._gridNeedsUpdate.set()

    def _onCancelBtnClicked(self, checked: bool):
        self._deleteAnyPendingGridTargets()
        self._targetComboBox.setCurrentIndex(-1)

    def _onFinishBtnClicked(self, checked: bool):
        self._pendingGridTargetKeys.clear()
        self._targetComboBox.setCurrentIndex(-1)


from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import json
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Navigator.Model.Session import Session, CoordinateSystems, CoordinateSystem
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util.pyvista import setActorUserTransform
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import transformToString, stringToTransform, concatenateTransforms, invertTransform
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.util.GUI.QLineEdit import QLineEditWithValidationFeedback
from NaviNIBS.util.GUI.QTableWidgetDragRows import QTableWidgetDragRows
from NaviNIBS.util.GUI.QValidators import OptionalTransformValidator
from NaviNIBS.util.pyvista.plotting import BackgroundPlotter

logger = logging.getLogger(__name__)


@attrs.define
class CoordinateSystemsPanel(MainViewPanel):
    _key: str = 'Set transforms'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-sync-outline'))

    # TODO


import attrs
from collections.abc import Sequence, Mapping
import contextlib
import logging
import typing as tp
from qtpy import QtCore, QtGui, QtWidgets

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem
from NaviNIBS.Navigator.Model import Session
from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)


K = tp.TypeVar('K', int, str)  # collection item key type
C = tp.TypeVar('C', bound=GenericCollection)  # collection type
CI = tp.TypeVar('CI', bound=GenericCollectionDictItem)  # collection item type


@attrs.define(slots=False, kw_only=True)
class CollectionTableModelBase(tp.Generic[K, C, CI]):
    _session: Session = attrs.field(repr=False)

    sigSelectionChanged: Signal = attrs.field(init=False, factory=lambda: Signal((list[K],)))  # emits which keys had selection changed (including both newly selected AND newly deselected keys)

    def __attrs_post_init__(self):
        pass


@attrs.define(slots=False)
class CollectionTableModel(CollectionTableModelBase[K, C, CI], QtCore.QAbstractTableModel):
    _columns: list[str] = attrs.field(factory=list)
    _attrColumns: list[str] = attrs.field(factory=list)
    _derivedColumns: dict[str, tp.Callable[[K], tp.Any]] = attrs.field(factory=dict)
    """
    Mapping from column key to function which generates derived value for given key/index
    """
    _boolColumns: list[str] = attrs.field(factory=list)
    """
    Columns (attr or derived) that should be represented by a checkbox
    """
    _decoratedColumns: list[str] = attrs.field(factory=list)
    """
    Columns (attr or derived) that include an icon. Values of referenced columns should be (QIcon, text) tuples. Usually, it will be necessary to set up a derived column for proper formatting.
    """
    _editableColumns: list[str] = attrs.field(factory=list)
    _columnLabels: dict[str, str] = attrs.field(factory=dict)
    """
    Mapping from column key to nice label; if a key is not included, will be used directly as a label
    """
    _editableColumnValidators: dict[str, tp.Callable[[K, tp.Any, tp.Any], bool]] = attrs.field(factory=dict)
    """
    Mapping from column key to validator function which returns True if passed (prevVal, newVal) is valid
    """
    _derivedColumnSetters: dict[str, tp.Callable[[K, str], None]] = attrs.field(factory=dict)
    """"
    Mapping from (editable, derived) column key to a setter function which, given a key/index 
    and a newly-edited string value, applies changes to the underlying collection.
    
    E.g. may be used for converting string text field to float for a numeric attribute.
    """
    _isSelectedAttr: tp.Optional[str] = None
    """
    Key of attr in collection indicating selection status; if None, selection state will not be synced to model; partially reliant on connected CollectionTableWidget to implement
    """

    _hasPlaceholderNewRow: bool = False
    """
    Whether to include a row at end of table as a placeholder for adding a new entry
    """
    _placeholderNewRowDefaults: dict[str, tp.Any] = attrs.field(factory=dict)
    """
    When including a row at end of table as a placeholder for adding a new entry, what defaults should be shown.
    This is a mapping from columnKey -> default value. If a column is not present in mapping, its default
    will be empty or false.
    """
    _addNewRowFromEditedPlaceholder: tp.Optional[tp.Callable[[tp.Any,...], K | None]] = attrs.field(default=None)
    """
    When user tries editing the placeholder row, a new row should be added to the collection. Specify how to
    create a new collection entry by providing a function that accepts as arguments any edited kwargs (e.g.
    a new key), creates an instance of a collection item, and adds this to the model, and returns the corresponding key.     
    
    Only used if `hasPlaceholderNewRow` is True. If not specified, columns in the placeholder row will not
    be editable even if they are editable in other rows.
    
    For example, using typical GenericCollection and GenericCollectionItem subclasses::
        
        def addNewRow(**kwargs) -> str:
            item = CollectionItem(**kwargs)
            collection.addItem(item)
            return item.key
            
        addNewRowFromEditedPlaceholder=addNewRow 
    """

    _collection: C = attrs.field(init=False, repr=False)

    _pendingChangeType: tp.Optional[str] = attrs.field(init=False, default=None)

    _lastSelectedKeys: list[K] = attrs.field(init=False, factory=list)  # only used if _isSelectedAttr is None

    def __attrs_post_init__(self):
        CollectionTableModelBase.__attrs_post_init__(self)
        QtCore.QAbstractTableModel.__init__(self)

        doAutoCollateColumns = len(self._columns) == 0

        for seq in (self._attrColumns,
                    self._derivedColumns.keys(),
                    self._boolColumns,
                    self._decoratedColumns,
                    self._editableColumns,
                    self._columnLabels.keys(),
                    self._editableColumnValidators.keys(),
                    self._derivedColumnSetters.keys(),
                    self._placeholderNewRowDefaults.keys()):
            for key in seq:
                if doAutoCollateColumns:
                    if key not in self._columns:
                        self._columns.append(key)
                else:
                    assert key in self._columns, f'{key} not in main columns list'

    @property
    def collection(self):
        return self._collection

    @property
    def collectionIsDict(self):
        return hasattr(self._collection, 'keys')

    @property
    def columns(self):
        """
        Note: result may be modified in place within modifyingColumns() context
        """
        return self._columns

    @columns.setter
    def columns(self, columns: list[str]):
        assert self._pendingChangeType == 'columns', 'Can only modify columns within a modifyingColumns() context'
        self._columns = columns
        # validation handled at end of modifyingColumns context

    @property
    def attrColumns(self):
        """
        Note: result may be modified in place within modifyingColumns() context
        """
        return self._attrColumns

    @attrColumns.setter
    def attrColumns(self, attrColumns: list[str]):
        assert self._pendingChangeType == 'columns', 'Can only modify columns within a modifyingColumns() context'
        self._attrColumns = attrColumns
        # validation handled at end of modifyingColumns context

    @property
    def derivedColumns(self):
        """
        Note: result may be modified in place within modifyingColumns() context
        """
        return self._derivedColumns

    @derivedColumns.setter
    def derivedColumns(self, derivedColumns: dict[str, tp.Callable[[K], tp.Any]]):
        assert self._pendingChangeType == 'columns', 'Can only modify columns within a modifyingColumns() context'
        self._derivedColumns = derivedColumns
        # validation handled at end of modifyingColumns context

    # TODO: write other column getter/setters in similar format (asserting modifyingColumns context)

    @contextlib.contextmanager
    def modifyingColumns(self):

        if self._pendingChangeType:
            logger.error(f'Previous change still pending: {self._pendingChangeType}')
            raise RuntimeError
        self._pendingChangeType = 'columns'
        self.layoutAboutToBeChanged.emit()

        yield

        unreferencedColumns = set(self._columns)
        for seq in (self._attrColumns,
                    self._derivedColumns.keys()):
            for key in seq:
                try:
                    unreferencedColumns.remove(key)
                except KeyError:
                    pass
        if len(unreferencedColumns) > 0:
            raise ValueError(f'Unreferenced columns: {unreferencedColumns}')

        for seq in (self._attrColumns,
                    self._derivedColumns.keys(),
                    self._boolColumns,
                    self._decoratedColumns,
                    self._editableColumns,
                    self._columnLabels.keys(),
                    self._editableColumnValidators.keys(),
                    self._derivedColumnSetters.keys(),
                    self._placeholderNewRowDefaults.keys()):
            for key in seq:
                assert key in self._columns, f'{key} not in main columns list'

        assert self._pendingChangeType == 'columns'
        self.layoutChanged.emit()
        self._pendingChangeType = None

        self._updateSelection(keys=None, attrKeys=None)

    def rowCount(self, parent: tp.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex]=...) -> int:
        return len(self._collection) + (1 if self._hasPlaceholderNewRow else 0)

    def columnCount(self, parent: tp.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex]=...) -> int:
        return len(self._columns)

    def flags(self, index:tp.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex]) -> QtCore.Qt.ItemFlags:
        colKey = self._columns[index.column()]
        flags = super().flags(index)
        if colKey in self._boolColumns and colKey in self._editableColumns:
            flags |= QtCore.Qt.ItemIsUserCheckable

        if colKey in self._editableColumns and colKey not in self._boolColumns:
            if self._hasPlaceholderNewRow and self._addNewRowFromEditedPlaceholder is None:
                # editing columns not allowed if in placeholder row
                itemKey = self.getCollectionItemKeyFromIndex(index.row())
                if itemKey is None:
                    # in placeholder row
                    pass  # don't set flag
                else:
                    flags |= QtCore.Qt.ItemIsEditable
            else:
                flags |= QtCore.Qt.ItemIsEditable

        return flags

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int=...) -> tp.Any:

        #logger.debug(f'{__name__} Getting headerData section {section} orientation {orientation} role {role}')

        match orientation:
            case QtCore.Qt.Horizontal:
                # column headers
                match role:
                    case QtCore.Qt.DisplayRole:
                        colKey = self._columns[section]
                        colLabel = self._columnLabels.get(colKey, colKey)
                        return colLabel
                    case _:
                        return None

            case QtCore.Qt.Vertical:
                # row headers
                match role:
                    case QtCore.Qt.DisplayRole:
                        return f'{section}'  # include row numbers
                    case _:
                        return None

            case _:
                raise NotImplementedError

    def data(self, index: tp.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex], role: int = ...) -> tp.Any:
        colKey = self._columns[index.column()]
        try:
            item = self.getCollectionItemFromIndex(index=index.row())
        except IndexError:
            return None  # TODO: double check what would be correct to return here

        if item is None and not self._hasPlaceholderNewRow:
            return None

        #logger.debug(f'Getting data for {self.getCollectionItemKeyFromIndex(index=index.row())} {colKey} role {role}')
        match role:
            case QtCore.Qt.DisplayRole | QtCore.Qt.ToolTipRole | QtCore.Qt.EditRole:
                if colKey in self._boolColumns:
                    # will be handled by CheckStateRole instead
                    return None
                elif colKey in self._attrColumns:
                    if item is None:
                        colVal = self._placeholderNewRowDefaults.get(colKey, (None, '') if colKey in self._decoratedColumns else '')
                    else:
                        colVal = getattr(item, colKey)
                    if colKey in self._decoratedColumns:
                        # assume val above is an (icon, text) tuple
                        assert len(colVal) == 2
                        colVal = colVal[1]
                    return str(colVal)
                elif colKey in self._derivedColumns:
                    colVal = self._derivedColumns[colKey](self.getCollectionItemKeyFromIndex(index=index.row()))
                    if colKey in self._decoratedColumns:
                        # assume val above is an (icon, text) tuple
                        assert len(colVal) == 2
                        colVal = colVal[1]
                    if isinstance(colVal, float):
                        colVal = f'{colVal:.4g}'  # TODO: make display precision (or format str) configurable rather than hardcoded
                    return str(colVal)
                else:
                    raise KeyError
            case QtCore.Qt.DecorationRole:
                if colKey in self._decoratedColumns:
                    if colKey in self._attrColumns:
                        if item is None:
                            colVal = self._placeholderNewRowDefaults.get(colKey, (None, ''))
                        else:
                            colVal = getattr(item, colKey)

                    elif colKey in self._derivedColumns:
                        colVal = self._derivedColumns[colKey](self.getCollectionItemKeyFromIndex(index=index.row()))

                    else:
                        raise KeyError

                    # assume val above is an (icon, text) tuple
                    assert len(colVal) == 2
                    colIcon = colVal[0]
                    assert colIcon is None or isinstance(colIcon, (QtGui.QColor, QtGui.QIcon, QtGui.QPixmap))
                    return colIcon

            case QtCore.Qt.CheckStateRole:
                if colKey in self._boolColumns:
                    if colKey in self._attrColumns:
                        if item is None:
                            colVal = self._placeholderNewRowDefaults.get(colKey, None)
                            if colVal is None:
                                return None  # don't show a checkbox for placeholder without a default
                        else:
                            colVal = getattr(item, colKey)
                    elif colKey in self._derivedColumns:
                        colVal = self._derivedColumns[colKey](self.getCollectionItemKeyFromIndex(index=index.row()))
                    else:
                        raise KeyError
                    assert isinstance(colVal, bool)
                    if colVal:
                        return int(QtCore.Qt.CheckState.Checked.value)
                    else:
                        return int(QtCore.Qt.CheckState.Unchecked.value)
                else:
                    return None
            case _:
                return None

    def setData(self, index:tp.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex], value:tp.Any, role: int = ...) -> bool:
        colKey = self._columns[index.column()]
        if colKey not in self._editableColumns:
            return False

        match role:
            case QtCore.Qt.EditRole:
                if colKey not in self._boolColumns:
                    logger.info(f'Editing column {colKey} to value {value}')
                    itemKey = self.getCollectionItemKeyFromIndex(index.row())

                    if colKey in self._editableColumnValidators:
                        oldValue = self.data(index, role)
                        isValid = self._editableColumnValidators[colKey](itemKey, oldValue, value)
                        if not isValid:
                            logger.warning('Attempted to set invalid value, rejecting change.')
                            return False

                    if self._hasPlaceholderNewRow and itemKey is None:
                        # tried to edit placeholder row, so create a new row and apply the edit to it
                        assert self._addNewRowFromEditedPlaceholder is not None
                        if (colKey in self._placeholderNewRowDefaults and value == self._placeholderNewRowDefaults[colKey]) \
                                or (colKey not in self._placeholderNewRowDefaults and len(value) == 0):
                            # no change from default, assume addition was cancelled
                            return False
                        newItemKey = self._addNewRowFromEditedPlaceholder(**{colKey: value})
                        if newItemKey is None:
                            # problem adding new row, reject change
                            return False
                        # TODO: change current item to newly created row
                        return True

                    if colKey in self._derivedColumnSetters:
                        self._derivedColumnSetters[colKey](itemKey, value)
                    elif colKey in self._attrColumns:
                        setattr(self._collection[itemKey], colKey, value)
                    else:
                        raise KeyError
                    return True
                else:
                    return False
            case QtCore.Qt.CheckStateRole:
                if colKey in self._boolColumns:
                    if colKey in self._attrColumns:
                        isChecked = value == QtCore.Qt.CheckState.Checked.value
                        itemKey = self.getCollectionItemKeyFromIndex(index.row())
                        if self._isSelectedAttr is None:
                            lastSelectedKeys = self._lastSelectedKeys
                        else:
                            lastSelectedKeys = [key for key in self._collection.keys() if getattr(self._collection[key], self._isSelectedAttr)]
                        if itemKey in lastSelectedKeys and True:
                            # apply edited checkbox value to all selected rows
                            self._collection.setAttribForItems(keys=lastSelectedKeys,
                                                               attribsAndValues={colKey: [isChecked] * len(lastSelectedKeys)})
                        else:
                            # apply edited checkbox value to current row only
                            logger.debug(f'Editing column {colKey} to value {isChecked}')
                            setattr(self.getCollectionItemFromIndex(index.row()), colKey, isChecked)
                            self.dataChanged.emit(index, index, [role])
                        return True
                    elif colKey in self._derivedColumns:
                        raise NotImplementedError
                    else:
                        raise KeyError
                else:
                    return False
            case _:
                return False

    def getCollectionItemFromIndex(self, index: int) -> tp.Optional[CI]:
        key = self.getCollectionItemKeyFromIndex(index)
        if key is None:
            return None  # return None as indicator of being in placeholder row

        return self._collection[key]

    def getCollectionItemKeyFromIndex(self, index: int) -> tp.Optional[K]:
        if index >= len(self._collection):
            if index == len(self._collection) and self._hasPlaceholderNewRow:
                return None  # return None as indicator of being in placeholder row
            elif index == 0:
                # this can sometimes happen after deleting all entries in a collection
                return None
            else:
                raise IndexError

        if self.collectionIsDict:
            key = list(self._collection.keys())[index]
            return key
        else:
            return index

    def getIndexFromCollectionItemKey(self, key: K) -> tp.Optional[int]:
        if self.collectionIsDict:
            try:
                index = list(self._collection.keys()).index(key)
            except ValueError:
                index = None
        else:
            if key < len(self._collection):
                index = key
            else:
                index = None
        return index

    def setWhichItemsSelected(self, selectedKeys: list[K]):
        """
        To be called by CollectionTableWidget to update model tracking of selection state.
        Does not directly change actual view selection state.
        """

        if self._pendingChangeType is not None:
            logger.debug(f'Ignoring setWhichItemsSelected ({selectedKeys}) since another change is pending ({self._pendingChangeType})')
            return  # ignore pending changes, assume will be handled later

        if self._isSelectedAttr is None:
            self._lastSelectedKeys = selectedKeys
        else:
            self._collection.setAttribForItems(self._collection.keys(),
                                               {self._isSelectedAttr:[key in selectedKeys for key in self._collection.keys()]})

    def _inferCollectionChangeType(self, keys: tp.Optional[list[K]], attrKeys: tp.Optional[list[str]]) -> str:

        changeType = None

        #return 'full'  # TODO: debug, delete

        if keys is None:
            changeType = 'full'
        else:
            indices = list(set(self.getIndexFromCollectionItemKey(key) for key in keys))
            if any(index is None for index in indices):
                if any(index is not None for index in indices):
                    # mix of previous items changing and new items being added; just do a full layout change
                    changeType = 'full'
                else:
                    # new items only
                    changeType = 'insertRows'
            else:
                # no new items
                # since we can't tell the difference at this point between a remove, a reorder, or a modify
                #  existing, just assume if all attributes are changing that we should do a full update
                #  (to handle any of these cases)
                # TODO: add a way to tell if we are going to do a remove and do more efficient partial removeRows instead
                if attrKeys is None:
                    changeType = 'full'
                else:
                    changeType = 'modifyExisting'


        return changeType

    def _onCollectionAboutToChange(self, keys: tp.Optional[list[K]], attrKeys: tp.Optional[list[str]] = None):
        """
        Should be called by subclass (e.g. via connected signal) whenever underlying collection is about to change
        :param keys: keys (or indices for non-dict collections) of items about to change; if None, all items are assumed to be about to change
        :param attrKeys: optional keys of attributes about to change; if None, all attributes are assumed to be about to change
        """

        if self._pendingChangeType:
            logger.error(f'Previous change still pending: {self._pendingChangeType}')
            raise RuntimeError
        self._pendingChangeType = self._inferCollectionChangeType(keys=keys, attrKeys=attrKeys)

        logger.debug(f'Signaling start of {self._pendingChangeType} change for keys {keys}, attrKeys {attrKeys}')

        match self._pendingChangeType:
            case 'full':
                self.layoutAboutToBeChanged.emit()

            case 'insertRows':
                indices = list(self.getIndexFromCollectionItemKey(key) for key in keys)
                assert all(index is None for index in indices)
                # assume being added at end of collection
                self.beginInsertRows(QtCore.QModelIndex(), len(self._collection), len(self._collection) + len(indices)-1)

            case 'modifyExisting':
                pass  # do nothing here (will emit dataChanged after change complete)

            case _:
                raise NotImplementedError

    def _onCollectionChanged(self, keys: tp.Optional[list[K]], attrKeys: tp.Optional[list[str]] = None):
        """
        Should be called by subclass (e.g. via connected signal) whenever underlying collection has changed
        :param keys: keys (or indices for non-dict collections) of items changed; if None, all items are assumed to be about to change
        :param attrKeys: optional keys of attributes changed; if None, all attributes are assumed to have changed
        """

        assert self._pendingChangeType is not None

        logger.debug(f'Signaling completion of {self._pendingChangeType} change for keys {keys}, attrKeys {attrKeys}')

        doUpdateSelection = False

        match self._pendingChangeType:
            case 'full':
                # TODO: update persistent indices with `changePersistentIndex` as required (?) by Qt model interface
                self.layoutChanged.emit()
                doUpdateSelection = True

            case 'insertRows':
                self.endInsertRows()

            case 'modifyExisting':
                indices = list(set(self.getIndexFromCollectionItemKey(key) for key in keys))
                assert all(index is not None for index in indices)
                assert len(indices) == len(keys), 'One or more keys repeated'
                minIndex = min(indices)
                maxIndex = max(indices)

                # TODO: also only emit for certain columns if attrKeys specified
                logger.debug(f'minIndex {minIndex}; maxIndex {maxIndex}, numColumns {len(self._columns)}')
                self.dataChanged.emit(self.index(minIndex, 0), self.index(maxIndex, len(self._columns)))

                doUpdateSelection = True

            case _:
                raise NotImplementedError

        self._pendingChangeType = None

        if doUpdateSelection:
            self._updateSelection(keys=keys, attrKeys=attrKeys)

        logger.debug(f'Done signaling completion for keys {keys}, attrKeys {attrKeys}')

    def _updateSelection(self, keys: list[str] | None, attrKeys: list[str] | None):
        logger.debug(f'Updating selection')
        if self._isSelectedAttr is not None and (attrKeys is None or self._isSelectedAttr in attrKeys):
            keysToEmit = keys
            if keysToEmit is None:
                if self.collectionIsDict:
                    keysToEmit = list(self._collection.keys())
                else:
                    keysToEmit = list(range(len(self._collection)))
            logger.debug(f'Keys with selection changing: {keysToEmit}')
            self.sigSelectionChanged.emit(keysToEmit)

    def getCollectionItemIsSelected(self, key: K) -> bool:
        if not self._isSelectedAttr:
            raise NotImplementedError

        return getattr(self._collection[key], self._isSelectedAttr)


@attrs.define(slots=False)
class FilteredCollectionModel(CollectionTableModelBase[K, C, CI], QtCore.QSortFilterProxyModel):
    """
    Base class for models that filter a collection model, e.g. Targets subset
    """

    _proxiedModel: CollectionTableModel = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        CollectionTableModelBase.__attrs_post_init__(self)
        QtCore.QSortFilterProxyModel.__init__(self)
        assert self._proxiedModel is not None, 'Should be set by subclass before calling super().__attrs_post_init__'
        self._proxiedModel.sigSelectionChanged.connect(self._onFullSelectionChanged, priority=-1)
        self.setSourceModel(self._proxiedModel)

    def filterAcceptsRow(self, sourceRow: int, sourceParent: QtCore.QModelIndex) -> bool:
        raise NotImplementedError  # should be implemented by subclass

    def filterAcceptsColumn(self, source_column: int, source_parent: QtCore.QModelIndex) -> bool:
        return True  # can be implemented by subclass to filter specific columns

    def getCollectionItemFromIndex(self, index: int) -> CI | None:
        return self._proxiedModel.getCollectionItemFromIndex(self.mapToSource(self.index(index, 0)).row())

    def getCollectionItemKeyFromIndex(self, index: int) -> str | None:
        return self._proxiedModel.getCollectionItemKeyFromIndex(self.mapToSource(self.index(index, 0)).row())

    def getIndexFromCollectionItemKey(self, key: str) -> int | None:
        proxiedRow = self._proxiedModel.getIndexFromCollectionItemKey(key)
        if proxiedRow is None:
            return None
        return self.mapFromSource(self._proxiedModel.index(proxiedRow, 0)).row()

    def getCollectionItemIsSelected(self, key: str) -> bool:
        return self._proxiedModel.getCollectionItemIsSelected(key)

    def setWhichItemsSelected(self, selectedKeys: list[K]):
        logger.debug(f'setWhichItemsSelected: {selectedKeys}')
        if True:
            # keep any keys in full model that were filtered out here still selected
            selectedKeys = set(selectedKeys)
            for sourceRow in range(self._proxiedModel.rowCount()):
                if not self.filterAcceptsRow(sourceRow=sourceRow, sourceParent=QtCore.QModelIndex()):
                    key = self._proxiedModel.getCollectionItemKeyFromIndex(sourceRow)
                    if self._proxiedModel.getCollectionItemIsSelected(key=key):
                        selectedKeys.add(key)

        self._proxiedModel.setWhichItemsSelected(selectedKeys)

    def _onFullSelectionChanged(self, keys: list[str]):
        logger.debug(f'onFullSelectionChanged: {keys}')
        self.sigSelectionChanged.emit(keys)  # TODO: filter to just subset of keys

